import inspect
from sortedcontainers import SortedList
import simpn.visualisation as vis


class SimVar:
    """
    A simulation variable SimVar has an identifier and a marking.
    A simulation variable can have multiple values. These values are available at particular times.
    For example, a variable van have the value 1 at time 0 (denoted as 1@0) and also 2@0.
    These values are also called tokens. Multiple tokens are called a marking of the variable.
    The marking is represented as a sorted list, which keeps tokens in the order of their priority, by default this is their timestamp.
    The functions put and remove, put and remove tokens in such a way that the order is maintained.

    :param _id: the identifier of the SimVar.
    :param priority: a function that takes a token as input and returns a value that is used to sort the tokens in the order in which they will be processed (lower values first). The default is processing in the order of the time of the token.
    """

    def __init__(self, _id, priority=lambda token: token.time):
        self._id = _id
        self.marking = SortedList(key=priority)
        self.checkpoints = dict()
        self.queue = SimVarQueue(self)
        self.visualize = True

    def put(self, value, time=0):
        """
        Put a token value in the SimVar at a particular time.

        :param value: the value to put in the SimVar.
        :param time: the time to make this value available at.
        """
        token = SimToken(value, time)
        self.add_token(token)

    def add_token(self, token, count=1):
        """
        Adds a token value in the SimVar.

        :param token: the token value to put in the SimVar.
        :param count: the number of times to put the token value in the SimVar (defaults to 1 time).
        """
        self.marking.add(token)

    def remove_token(self, token):
        """
        Removes a token value (once) from the SimVar.

        :param token: the token value to remove from the SimVar.
        """
        if token in self.marking:
            self.marking.remove(token)
        else:
            raise LookupError("No token '" + token + "' at place '" + str(self) + "'.")          
    
    def get_id(self):
        return self._id
    
    def __str__(self):
        return self._id

    def __repr__(self):
        return self.__str__()

    def store_checkpoint(self, name):
        """
        Stores a checkpoint of the SimVar marking with the given name. The checkpoint can be restored later with restore_checkpoint.
        """
        self.checkpoints[name] = [token.copy() for token in self.marking]
    
    def restore_checkpoint(self, name):
        """
        Restores the SimVar marking from the checkpoint with the given name.
        """
        if name in self.checkpoints:
            self.marking.clear()
            for token in self.checkpoints[name]:
                self.add_token(token)
        else:
            raise LookupError("No checkpoint '" + name + "' at place '" + str(self) + "'.")

    def get_visualisation(self):
        return vis.PlaceViz(self)

    def set_invisible(self):
        self.visualize = False

class SimVarQueue(SimVar):
    """
    A simulation variable that contains the queue of tokens from another SimVar.
    The identifier of this SimVar is <simvar_id>.queue, where simvar_id is the _id of the SimVar of which it contains the queue.
    Regular SimVar have a queue property that refers to their SimVarQueue variable.
    """
    QUEUE_SUFFIX = ".queue"
    
    def __init__(self, simvar):
        self._id = simvar._id + SimVarQueue.QUEUE_SUFFIX
        self.simvar = simvar

    @property
    def marking(self):
        token = SimToken(list(self.simvar.marking), 0)
        return [token]

    def put(self, value, time=0):
        pass

    def add_token(self, token, count=1):
        # adding a queue token means adding an entire queue.
        if count != 1:
            raise TypeError(self._id + ": the queue is placed back multiple times (count != 1). However, there can only be one queue.")
        # we check if the token is a list of tokens (i.e. a queue) and if so, we add all tokens in the queue.
        try:
            for t in token:
                if not isinstance(t, SimToken):
                    raise TypeError(self._id + ": something went wrong placing the queue back with value " + str(token) + ". Element " + str(t) + " does not appear to be a token, but the queue must be a list of tokens. Maybe you just added a value, not a token?")
                self.simvar.add_token(t)
        except:
            raise TypeError(self._id + ": something went wrong placing the queue back with value " + str(token) + ".")

    def remove_token(self, token):
        # removing a queue token means removing the queue.
        # However, the queue object should remain intact, because that is the same object that is passed as a token.
        # Therefore, we set the marking to a new empty list.
        self.simvar.marking = SortedList(key=self.simvar.marking.key)

    def __str__(self):
        return self._id

    def __repr__(self):
        return self.__str__()	
    

class SimVarTime(SimVar):
    """
    A simulation variable that contains the simulation time.
    This variable only always has a single token value, which is the simulation time.
    The identifier of this SimVar is 'time', which consequently is a reserved name.
    """
    TIME_ID = "time"
    
    def __init__(self, problem):
        self._id = SimVarTime.TIME_ID
        self.checkpoints = dict()
        self.problem = problem
        self.visualize = True

    @property
    def marking(self):
        token = SimToken(self.problem.clock)
        return SortedList([token])

    def put(self, value, time=0):
        pass

    def add_token(self, token, count=1):
        pass

    def remove_token(self, token):
        pass

    def __str__(self):
        return self._id

    def __repr__(self):
        return self.__str__()


class SimEvent:
    """
    A simulation event SimEvent that can happen when tokens are available on all of its
    incoming SimVar and its guard evaluates to True. When it happens, it consumes a token from each
    of its incoming SimVar and produces a token on each of its outgoing places according to `behavior`.
    The behavior takes just the values of the SimVar as input, but produces tokens as output: values with a delay.
    The delay can also be 0, meaning the token is produced with 0 delay.
    For example, consider the event with `incoming` `SimVar` [a, b] that have 2@1 and 2@1.
    The event has `outgoing` `SimVar` [c, d], behavior lambda a, b: [(a + b)@+1, (a - b)@+2].
    This event can happen for a = 2@1 and b = 2@1. Thus, it will happen at time 1, generating
    [3@+1, 1@+2] according to its behavior. Therefore, after the event has happened, c and d
    will have the tokens 3@2 and 1@3.

    :param _id: the identifier of the event.
    :param incoming: a list of incoming SimVar of the event.
    :param outgoing: a list of outgoing SimVar of the event
    :param guard: a function that takes as many parameters as there are incoming SimVar. The function must evaluate to True or False for all possible values of SimVar. The event can only happen for values for which the guard function evaluates to True.
    :param behavior: a function that takes as many parameters as there are incoming SimVar. The function must return a list with as many elements as there are outgoing SimVar. The elements must be tokens that carry the resulting values and the delay with which these values become available. When the event happens, the function is performed on the incoming SimVar and the result of the function is put on the outgoing SimVar with the corresponding delays.
    """

    def __init__(self, _id, guard=None, behavior=None, incoming=None, outgoing=None):
        self._id = _id
        self.guard = guard
        if incoming is None:
            self.incoming = []
        else:
            self.incoming = incoming
        if outgoing is None:
            self.outgoing = []
        else:
            self.outgoing = outgoing
        self.behavior = behavior
        self.visualize = True

    def set_guard(self, func):
        """
        Set the guard function.

        :param func: a function with as many input parameters as there are incoming SimVar, which generates a list with as many elements as there are outgoing SimVar.
        """
        self.guard = func

    def set_inflow(self, incoming):
        """
        Set the incoming SimVar.

        :param incoming: a list of SimVar.
        """
        self.incoming = incoming

    def set_outflow(self, outgoing):
        """
        Set the outgoing SimVar.

        :param outgoing: a list of SimVar.
        """
        self.outgoing = outgoing

    def set_behavior(self, behavior):
        """
        Set the behavior.

        :param behavior: the behavior to set.
        """
        self.behavior = behavior

    def get_id(self):
        return self._id

    def __str__(self):
        return self._id

    def __repr__(self):
        return self.__str__()

    def get_visualisation(self):
        return vis.TransitionViz(self)

    def set_invisible(self):
        self.visualize = False

class SimToken:
    """
    A token SimToken, which is a possible value of a SimVar. A token has a value and the time at which this value is available in a SimVar.
    When the SimToken is used as the return value of an event behavior, the delay can be used. This represents the delay with which the value will be available.
    The value will then be available at <time of event> + <delay>.

    :param value: the value of the token.
    :param time: the time at which the value is available. 
    :param delay: should only be used when returning the token from an event behavior. The value represents the delay. The token will then be available as <time of event> + <delay>.

    """
    def __init__(self, value, time=0, delay=0):
        self.value = value
        self.time = time
        self.delay = delay

    def __eq__(self, token):
        return self.value == token.value and self.time == token.time

    def __lt__(self, token):
        if isinstance(token.value, dict) and isinstance(self.value, dict):
            return self.time < token.time or (self.time == token.time and list(self.value.values()) < list(token.value.values()))
        elif isinstance(token.value, (str, int, float, complex, bool)) and isinstance(self.value, (str, int, float, complex, bool)):
            return self.time < token.time or (self.time == token.time and self.value < token.value)
        elif isinstance(token.value, (list, tuple)) and isinstance(self.value, (list, tuple)):
            return self.time < token.time or (self.time == token.time and str(self.value) < str(token.value)) #relied on string just to impose a strict ordering
        elif isinstance(self.value, dict) and isinstance(token.value, (str, int, float, complex, bool, list, tuple)):
            return self.time < token.time or (self.time == token.time and list(self.value.values()) < [token.value])
        elif isinstance(self.value, (str, int, float, complex, bool, list, tuple)) and isinstance(token.value, dict):
            return self.time < token.time or (self.time == token.time and [self.value] < list(token.value.values()))
        elif isinstance(self.value, (list, tuple)) and isinstance(token.value, (str, int, float, complex, bool)):
            return self.time < token.time or (self.time == token.time and self.value < [token.value])
        elif isinstance(self.value, (str, int, float, complex, bool)) and isinstance(token.value, (list, tuple)):
            return self.time < token.time or (self.time == token.time and [self.value] < token.value)
        else:
            raise TypeError("Unsupported comparison between '{}' and '{}'".format(type(self.value), type(token.value)))

    def __hash__(self):
        return (self.value, self.time).__hash__()

    def __str__(self):
        result = str(self.value)
        if self.time is not None:
            result += "@" + str(self.time)
        return result

    def __repr__(self):
        return self.__str__()
    
    def copy(self):
        return SimToken(self.value, self.time)


class SimProblem:
    """
    A simulation problem SimProblem, which consists of a collection of simulation variables SimVar and a collection of simulation events SimEvent.
    The simulation has a time and a marking of SimVar. A marking is an assignment of values to SimVar variables (also see SimVar).
    The difference between a normal Python variable and a SimVar is that: (1) a SimVar can have multiple values; and (2) these values have a simulation time
    at which they become available. A SimEvent is a function that takes SimVar values as input and produces SimVar values (also see SimEvent). The difference with normal
    Python functions is that events can happen whenever its incoming SimVar parameters have a value. If a SimVar parameter has multiple values,
    it can happen on any of these values.

    :param debugging: if set to True, produces more information for debugging purposes (defaults to True).
    :param binding_priority: a function that takes a list of binding as input and returns the binding that will be selected in case there are multiple possible bindings to fire.
    """
    def __init__(self, debugging=True, binding_priority=lambda bindings: bindings[0]):
        self.places = []
        self.events = []
        self.prototypes = []
        self.id2node = dict()
        self.clock = 0
        self._debugging = debugging
        self.clock_checkpoints = dict()
        self.binding_priority = binding_priority

    def __str__(self):
        result = ""
        result += "P={"
        for i in range(len(self.places)):
            result += str(self.places[i])
            if i < len(self.places) - 1:
                result += ","
        result += "}\n"
        result += "T={"
        for i in range(len(self.events)):
            result += str(self.events[i])
            if self.events[i].guard is not None:
                result += "[" + self.events[i].guard.__name__ + "]"
            result += ":" + self.events[i].behavior.__name__
            if i < len(self.events) - 1:
                result += ","
        result += "}\n"
        arcs = []
        for t in self.events:
            for p in t.incoming:
                arcs.append("(" + str(p) + "," + str(t) + ")")
            for p in t.incoming:
                arcs.append("(" + str(t) + "," + str(p) + ")")
        result += "A={"
        for i in range(len(arcs)):
            result += arcs[i]
            if i < len(arcs) - 1:
                result += ","
        result += "}\n"
        markings = []
        for p in self.places:
            if len(p.marking) > 0:
                mstr = str(p) + ":["
                ti = 0
                for token in p.marking:
                    mstr += str(token)
                    if ti < len(p.marking) - 1:
                        mstr += ", "
                    ti += 1
                markings.append(mstr + "]")
        result += "M={"
        for i in range(len(markings)):
            result += markings[i]
            if i < len(markings) - 1:
                result += ","
        result += "}"                
        return result

    def add_var(self, name, priority=lambda token: token.time):
        """
        Creates a new SimVar with the specified name as identifier. Adds the SimVar to the problem and returns it.

        :param name: a name for the SimVar.
        :param priority: a function that takes a token as input and returns a value that is used to sort the tokens in the order in which they will be processed (lower values first). The default is processing in the order of the time of the token.
        :return: a SimVar with the specified name as identifier.
        """
        # Generate and add SimVar
        result = SimVar(name, priority=priority)
        self.add_prototype_var(result)
        return result
    def add_place(*args, **kwargs):
        """
        Adds a SimVar to the problem. This function is a wrapper around add_var for people with a Petri-net background.
        """
        self = args[0]
        return self.add_var(*(args[1:]), **kwargs)

    def add_prototype_var(self, var):
        """
        Adds the SimVar to the problem. This function should only be used for prototypes.
        """
        # Check name
        name = var.get_id()
        if name in self.id2node:
            raise TypeError("Node with name " + name + " already exists. Names must be unique.")
        if name.endswith(SimVarQueue.QUEUE_SUFFIX):
            raise TypeError("Cannot create SimVar with name " + name + ". Names ending with " + SimVarQueue.QUEUE_SUFFIX + " are reserved for SimVar queues. They are automatically generated.")
        if name == SimVarTime.TIME_ID:
            raise TypeError("Cannot create SimVar with name " + name + ". " + SimVarTime.TIME_ID + " is reserved for the time variable. If you just want to get the time variable, use the .var() method instead.")

        # Add SimVar
        self.places.append(var)
        self.id2node[name] = var

    def add_prototype_event(self, event):
        """
        Adds the SimEvent to the problem. This function should only be used for prototypes.        
        """
        # Check name
        name = event.get_id()
        if name in self.id2node:
            raise TypeError("Node with name " + name + " already exists. Names must be unique.")

        # Add SimEvent
        self.events.append(event)
        self.id2node[name] = event


    def var(self, name):
        """
        Returns the SimVar with the given name.
        Raises an error if no such SimVar exists.

        :param name: the name of the SimVar.
        :return: the SimVar with the given name or an Error.
        """
        if name in self.id2node:
            if isinstance(self.id2node[name], SimVar):
                return self.id2node[name]
            else:
                raise TypeError(name + " is not a SimVar.")
        elif name == SimVarTime.TIME_ID:
            # Generate and add SimVarTime
            time_var = SimVarTime(self)
            self.places.append(time_var)
            self.id2node[time_var._id] = time_var
            return time_var
        else:
            raise LookupError("SimVar " + name + ": does not exist.")
    def place(*args, **kwargs):
        """
        Returns the SimVar with the given name. This function is a wrapper around var for people with a Petri-net background.
        """
        self = args[0]
        return self.var(*(args[1:]), **kwargs)

    def store_checkpoint(self, name):
        """
        Stores the state of the simulator as a checkpoint with the given name. The checkpoint can be restored later with restore_checkpoint.
        """
        for p in self.places:
            p.store_checkpoint(name)
        self.clock_checkpoints[name] = self.clock

    def restore_checkpoint(self, name):
        """
        Restores the state of the simulator to the checkpoint with the given name.
        """
        if name not in self.clock_checkpoints:
            raise LookupError("No checkpoint '" + name + "'.")
        for p in self.places:
            p.restore_checkpoint(name)
        self.clock = self.clock_checkpoints[name]

    def add_event(self, inflow, outflow, behavior, name=None, guard=None):
        """
        Creates a new SimEvent with the specified parameters (also see SimEvent). Adds the SimEvent to the problem and returns it.

        :param inflow: a list of incoming SimVar of the event.
        :param outflow: a list of outgoing SimVar of the event/
        :param behavior: a function that takes as many parameters as there are incoming SimVar. The function must return a list with as many elements as there are outgoing SimVar. When the event happens, the function is performed on the incoming SimVar and the result of the function is put on the outgoing SimVar.
        :param name: the identifier of the event.
        :param guard: a function that takes as many parameters as there are incoming SimVar. The function must evaluate to True or False for all possible values of SimVar. The event can only happen for values for which the guard function evaluates to True.
        :return: a SimEvent with the specified parameters.
        """

        # Check name
        t_name = name
        if t_name is None:
            if behavior.__name__ == "<lambda>":
                raise TypeError("Event name must be set or procedure behavior function must be named.")
            else:
                t_name = behavior.__name__
        if t_name in self.id2node:
            raise TypeError("Event " + t_name + ": node with the same name already exists. Names must be unique.")

        # Check inflow
        c = 0
        for i in inflow:
            if not isinstance(i, SimVar):
                raise TypeError("Event " + t_name + ": inflow with index " + str(c) + " is not a SimVar.")
            c += 1

        # Check outflow
        c = 0
        for o in outflow:
            if not isinstance(o, SimVar):
                raise TypeError("Event " + t_name + ": outflow with index " + str(o) + " is not a SimVar.")
            c += 1

        # Check behavior function
        if not callable(behavior):
            raise TypeError("Event " + t_name + ": the behavior must be a function. (Maybe you made it a function call, exclude the brackets.)")
        parameters = inspect.signature(behavior).parameters
        num_mandatory_params = sum(
            1 for p in parameters.values()
            if p.kind not in [inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD] # count all parameters which are not "*args" or "**kwargs"
        )
        if num_mandatory_params > len(inflow):
            raise TypeError("Event " + t_name + ": the behavior function must take as many parameters as there are input variables.")

        # Check constraint
        if guard is not None:
            if not callable(guard):
                raise TypeError("Event " + t_name + ": the constraint must be a function. (Maybe you made it a function call, exclude the brackets.)")
            parameters = inspect.signature(guard).parameters
            num_mandatory_params = sum(
                1 for p in parameters.values()
                if p.kind not in [inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD] # count all parameters which are not "*args" or "**kwargs"
            )
            if num_mandatory_params > len(inflow):
                raise TypeError("Event " + t_name + ": the constraint function must take as many parameters as there are input variables.")

        # Check queue operations: if the inflow is only queue vars, but the outflow is not, we post a warning.
        inflow_queues = all([isinstance(i, SimVarQueue) for i in inflow])
        outflow_queues = all([isinstance(o, SimVarQueue) for o in outflow])
        if inflow_queues and not outflow_queues:
            print("WARNING: Event " + t_name + ": has only queue variables in the inflow, but not in the outflow. This may lead to unexpected behavior, because queues do not have time, so the outflow tokens are produced at time=0.")

        # Generate and add SimEvent
        result = SimEvent(t_name, guard=guard, behavior=behavior, incoming=inflow, outgoing=outflow)
        self.events.append(result)
        self.id2node[t_name] = result

        return result
    def add_transition(*args, **kwargs):
        """
        Creates a new SimEvent with the specified parameters. This function is a wrapper around add_event for people with a Petri-net background.
        """
        self = args[0]
        return self.add_event(*(args[1:]), **kwargs)

    def add_prototype(self, prototype):
        if len(prototype.places) == 0 and len(prototype.events) == 0:
            raise TypeError("Prototype " + prototype.name + ": does not have any variables or events. That should not be possible, please notify the developer.")
        for place in prototype.places:
            if place not in self.places:
                raise TypeError("Prototype " + prototype.name + ": did not add all its places to the model. That should not be possible, please notify the developer.")
        for event in prototype.events:
            if event not in self.events:
                raise TypeError("Prototype " + prototype.name + ": did not add all its events to the model. That should not be possible, please notify the developer.")
        if prototype.get_id() in self.id2node:
            raise TypeError("Prototype " + prototype.name + ": node with the same name already exists. Names must be unique.")
        self.prototypes.append(prototype)

    def set_binding_priority(self, func):
        """
        Sets the binding priority function.

        :param func: a function that takes a list of binding as input and returns the binding that will be selected in case there are multiple possible bindings to fire.
        """
        self.binding_priority = func

    @staticmethod
    def tokens_combinations(event):
        """
        Creates a list of token combinations that are available to the specified event.
        These are combinations of tokens that are on the incoming SimVar of the event.
        For example, if a event has incoming SimVar a and b with tokens 1@0 on a and 2@0, 3@0 on b,
        the possible combinations are [(a, 1@0), (b, 2@0)] and [(a, 1@0), (b, 3@0)]

        :param event: the event to return the token combinations for.
        :return: a list of lists, each of which is a token combination.
        """
        # create all possible combinations of incoming token values
        bindings = [[]]
        for place in event.incoming:
            new_bindings = []
            for token in place.marking:  # get set of colors in incoming place
                for binding in bindings:
                    new_binding = binding.copy()
                    new_binding.append((place, token))
                    new_bindings.append(new_binding)
            bindings = new_bindings
        return bindings

    def event_bindings(self, event):
        """
        Calculates the set of bindings that enables the given event.
        Each binding is a tuple ([(place, token), (place, token), ...], time) that represents a single enabling binding.
        A binding is
        a possible token combination (see token_combinations), for which the event's
        guard function evaluates to True. In case there is no guard function, any combination is also a binding.
        The time is the time at which the latest token is available.
        For example, if a event has incoming SimVar a and b with tokens 1@2 on a and 2@3, 3@1 on b,
        the possible bindings are ([(a, 1@2), (b, 2@3)], 3) and ([(a, 1@2), (b, 3@1)], 2)

        :param event: the event for which to calculate the enabling bindings.
        :return: list of tuples ([(place, token), (place, token), ...], time)
        """
        if len(event.incoming) == 0:
            raise Exception("Though it is strictly speaking possible, we do not allow events like '" + str(self) + "' without incoming arcs.")

        bindings = self.tokens_combinations(event)

        # a binding must have all incoming places
        nr_incoming_places = len(event.incoming)
        new_bindings = []
        for binding in bindings:
            if len(binding) == nr_incoming_places:
                new_bindings.append(binding)
        bindings = new_bindings

        # if a event has a guard, only bindings are enabled for which the guard evaluates to True
        result = []
        for binding in bindings:
            variable_values = []
            time = None
            for (place, token) in binding:
                variable_values.append(token.value)
                if time is None or token.time > time:
                    time = token.time
            enabled = True
            if event.guard is not None:
                try:
                    enabled = event.guard(*variable_values)
                except Exception as e:
                    raise TypeError("Event " + event + ": guard generates exception for values " + str(variable_values) + ".") from e
                if self._debugging and not isinstance(enabled, bool):
                    raise TypeError("Event " + event + ": guard does evaluate to a Boolean for values " + str(variable_values) + ".")
            if enabled:
                result.append((binding, time))
        return result

    def bindings(self):
        """
        Calculates the set of timed bindings that is enabled over all events in the problem.
        Each binding is a tuple ([(place, token), (place, token), ...], time, event) that represents a single enabling binding.
        If no timed binding is enabled at the current clock time, updates the current clock time to the earliest time at which there is.
        :return: list of tuples ([(place, token), (place, token), ...], time, event)
        """
        timed_bindings = []
        min_enabling_time = None
        for t in self.events:
            for (binding, time) in self.event_bindings(t):
                timed_bindings.append((binding, time, t))
                if min_enabling_time is None or time < min_enabling_time:
                    min_enabling_time = time
        # timed bindings are only enabled if they have time <= clock
        # if there are no such bindings, set the clock to the earliest time at which there are
        if min_enabling_time is not None and min_enabling_time > self.clock:
            self.clock = min_enabling_time
            # We now also need to update the bindings, because the SimVarTime may have changed and needs to be updated.
            # TODO This is inefficient, because we are recalculating all bindings, while we only need to recalculate the ones that have SimVarTime in their inflow.
            timed_bindings = [] 
            for t in self.events:
                for (binding, time) in self.event_bindings(t):
                    timed_bindings.append((binding, time, t))

        # now return the untimed bindings + the timed bindings that have time <= clock
        return [(binding, time, t) for (binding, time, t) in timed_bindings if time <= self.clock]

    def fire(self, timed_binding):
        """
        Fires the specified timed binding.
        Binding is a tuple ([(place, token), (place, token), ...], time, event)
        """
        (binding, time, event) = timed_binding

        # process incoming places:
        variable_assignment = []
        for (place, token) in binding:
            # remove tokens from incoming places
            place.remove_token(token)
            # assign values to the variables on the arcs
            variable_assignment.append(token.value)

        # calculate the result of the behavior of the event
        try:
            result = event.behavior(*variable_assignment)
        except Exception as e:
            raise TypeError("Event " + str(event) + ": behavior function generates exception for values " + str(variable_assignment) + ".") from e
        if self._debugging:
            if type(result) != list:
                raise TypeError("Event " + str(event) + ": behavior function does not generate a list for values " + str(variable_assignment) + ".")
            if len(result) != len(event.outgoing):
                raise TypeError("Event " + str(event) + ": behavior function does not generate as many values as there are output variables for values " + str(variable_assignment) + ".")
            i = 0
            for r in result:
                if r is not None:
                    if isinstance(event.outgoing[i], SimVarQueue):
                        if not isinstance(r, list):
                            raise TypeError("Event " + str(event) + ": does not generate a queue for variable " + str(event.outgoing[i]) + " for values " + str(variable_assignment) + ".")
                    else:
                        if not isinstance(r, SimToken):
                            raise TypeError("Event " + str(event) + ": does not generate a token for variable " + str(event.outgoing[i]) + " for values " + str(variable_assignment) + ".")
                        if not (type(r.delay) is int or type(r.delay) is float):
                            raise TypeError("Event " + str(event) + ": does not generate a numeric value for the delay of variable " + str(event.outgoing[i]) + " for values " + str(variable_assignment) + ".")
                        if not (type(r.time) is int or type(r.time) is float):
                            raise TypeError("Event " + str(event) + ": does not generate a numeric value for the time of variable " + str(event.outgoing[i]) + " for values " + str(variable_assignment) + ".")
                i += 1

        for i in range(len(result)):
            if result[i] is not None:
                if isinstance(event.outgoing[i], SimVarQueue):
                    event.outgoing[i].add_token(result[i])
                else:
                    if result[i].time > 0 and result[i].delay == 0:
                        raise TypeError("Deprecated functionality: Event " + str(event) + ": generates a token with a delay of 0, but a time > 0, for variable " + str(event.outgoing[i]) + " for values " + str(variable_assignment) + ". It seems you are using the time of the token to represent the delay.")
                    token = SimToken(result[i].value, time=self.clock + result[i].delay)
                    event.outgoing[i].add_token(token)

    def step(self):
        """
        Executes a single step of the simulation.
        If multiple events can happen, one is selected at random.
        Returns the binding that happened, or None if no event could happen.
        """
        bindings = self.bindings()
        if len(bindings) > 0:
            timed_binding = self.binding_priority(bindings)
            self.fire(timed_binding)
            return timed_binding
        return None

    def simulate(self, duration, reporter=None):
        """
        Executes a simulation run for the problem for the specified duration.
        A simulation run executes events until this is no longer possible, or until the specified duration.
        If at any moment multiple events can happen, one is selected at random.
        At the end of the simulation run, the problem is in the state that is the result of the run.
        If the reporter is set, its callback function is called with parameters (binding, time, event)
        each time a event happens (for a binding at a moment in simulation time).

        :param duration: the duration of the simulation.
        :param reporter: a class that implements a callback function, which is called each time a event happens. reporter can also be a list of reporters, in which case the callback function of each reporter is called.
        """
        active_model = True
        while self.clock <= duration and active_model:
            bindings = self.bindings()
            if len(bindings) > 0:
                timed_binding = self.binding_priority(bindings)
                self.fire(timed_binding)
                if reporter is not None:
                    if type(reporter) == list:
                        for r in reporter:
                            r.callback(timed_binding)
                    else:
                        reporter.callback(timed_binding)
            else:
                active_model = False


def event(sim_problem: SimProblem, inflow: list, outflow: list, guard=None):
    """
    A decorator that can be used to turn a Python function into a SimEvent.
    The event will be added to the specified sim_problem.
    If they do not yet exist, SimVar will be added for each parameter of the Python function and for each element in outflow.
    The parameters of the function will be treated as inflow variables of the SimEvent (by name).
    The elements of outflow must be names of SimVar and will be treated as outflow variables of the SimEvent (by name).
    The name of the SimEvent will be the name of the Python function.

    :param sim_problem: the SimProblem to which to add the SimEvent.
    :param inflow: a list of names of inflow variables. Just there for documentation purposes, it is not being used.
    :param outflow: a list of names of outflow variables.
    :param guard: an optional guard for the SimEvent.
    :return: the Python function.
    """
    def decorator_event(func):
        parameters = inspect.signature(func).parameters
        for parameter in list(parameters.keys()) + outflow:
            if parameter not in sim_problem.id2node:
                sim_problem.add_var(parameter)
        inflow_svars = [sim_problem.id2node[parameter] for parameter in parameters]
        outflow_svars = [sim_problem.id2node[parameter] for parameter in outflow]
        sim_problem.add_event(inflow_svars, outflow_svars, func, guard=guard)
        return func
    return decorator_event



import igraph
import os
import traceback
os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = "hide"
import pygame
import simpn.assets as assets
from enum import Enum, auto


MAX_SIZE = 1920, 1080
# colors
TUE_RED = (200, 25, 25)
TUE_LIGHTRED = (249, 204, 204)
TUE_BLUE = (16, 16, 115)
TUE_LIGHTBLUE = (188, 188, 246)
TUE_GREY = (242, 242, 242)
WHITE = (255, 255, 255)
# sizes
STANDARD_NODE_WIDTH, STANDARD_NODE_HEIGHT = 50, 50
NODE_SPACING = 100
GRID_SPACING = 50
LINE_WIDTH = 2
ARROW_WIDTH, ARROW_HEIGHT = 12, 10
TEXT_SIZE = 16


class Shape(Enum):
    TRANSTION = auto()
    PLACE = auto()
    EDGE = auto()


class Hook(Enum):
    LEFT = auto()
    RIGHT = auto()
    TOP = auto()
    BOTTOM = auto()


class Edge:
    # An edge from a start node to and end node
    # start is a pair (node, Hook), where node is start node of the flow and Hook is the hook on that node
    # end is a pair (node, Hook), where node is start node of the flow and Hook is the hook on that node
    # intermediate is a list of pairs (x_new, y_new) of intermediate points on the flow, such that the flow
    #   is routed through these intermediate points. An intermediate point is computed from the previous point (x, y)
    #   by using (x_new if x_new >= 0 else x, y_new if y_new >= 0 else y)
    def __init__(self, start, end):
        self._start = start
        self._end = end

    def draw(self, screen):
        start_node_xy = self.get_start_node().get_pos()
        start_node_width, start_node_height = self.get_start_node()._width, self.get_start_node()._height
        end_node_xy = self.get_end_node().get_pos()
        end_node_width, end_node_height = self.get_end_node()._width, self.get_end_node()._height
        if start_node_xy[1] - start_node_height/2 > end_node_xy[1] + end_node_height/2:
            self.set_start_hook(Hook.TOP)
        elif start_node_xy[1] + start_node_height/2 < end_node_xy[1] - end_node_height/2:
            self.set_start_hook(Hook.BOTTOM)
        elif start_node_xy[0] - start_node_width/2 > end_node_xy[0] + end_node_width/2:
            self.set_start_hook(Hook.LEFT)
        else:
            self.set_start_hook(Hook.RIGHT)
        if end_node_xy[1] - end_node_height/2 > start_node_xy[1] + start_node_height/2:
            self.set_end_hook(Hook.TOP)
        elif end_node_xy[1] + end_node_height/2 < start_node_xy[1] - start_node_height/2:
            self.set_end_hook(Hook.BOTTOM)
        elif end_node_xy[0] - end_node_width/2 > start_node_xy[0] + start_node_width/2:
            self.set_end_hook(Hook.LEFT)
        else:
            self.set_end_hook(Hook.RIGHT)
        start = pygame.Vector2(self._start[0].hook(self._start[1]))
        end = pygame.Vector2(self._end[0].hook(self._end[1]))
        arrow = start - end
        angle = arrow.angle_to(pygame.Vector2(0, -1))
        body_length = arrow.length() - ARROW_HEIGHT

        # if the end node does not want to show arrowheads, we simply draw a line from start to end
        if not self.get_end_node()._show_arrowheads:
            pygame.draw.line(screen, TUE_BLUE, start, end, int(LINE_WIDTH*1.5))
            return

        # Create the triangle head around the origin
        head_verts = [
            pygame.Vector2(0, ARROW_HEIGHT / 2),  # Center
            pygame.Vector2(ARROW_WIDTH / 2, -ARROW_HEIGHT / 2),  # Bottomright
            pygame.Vector2(-ARROW_WIDTH / 2, -ARROW_HEIGHT / 2),  # Bottomleft
        ]
        # Rotate and translate the head into place
        translation = pygame.Vector2(0, arrow.length() - (ARROW_HEIGHT / 2)).rotate(-angle)
        for i in range(len(head_verts)):
            head_verts[i].rotate_ip(-angle)
            head_verts[i] += translation
            head_verts[i] += start

        pygame.draw.polygon(screen, TUE_BLUE, head_verts)

        # Stop weird shapes when the arrow is shorter than arrow head
        if arrow.length() >= ARROW_HEIGHT:
            # Calculate the body rect, rotate and translate into place
            body_verts = [
                pygame.Vector2(-LINE_WIDTH / 2, body_length / 2),  # Topleft
                pygame.Vector2(LINE_WIDTH / 2, body_length / 2),  # Topright
                pygame.Vector2(LINE_WIDTH / 2, -body_length / 2),  # Bottomright
                pygame.Vector2(-LINE_WIDTH / 2, -body_length / 2),  # Bottomleft
            ]
            translation = pygame.Vector2(0, body_length / 2).rotate(-angle)
            for i in range(len(body_verts)):
                body_verts[i].rotate_ip(-angle)
                body_verts[i] += translation
                body_verts[i] += start

            pygame.draw.polygon(screen, TUE_BLUE, body_verts)

    def get_shape(self):
        return Shape.EDGE

    def get_start_node(self):
        return self._start[0]
    
    def get_end_node(self):
        return self._end[0]

    def set_start_hook(self, hook):
        self._start = (self._start[0], hook)
    
    def set_end_hook(self, hook):
        self._end = (self._end[0], hook)


class Node:
    def __init__(self, model_node):
        self._model_node = model_node
        self._pos = (0, 0)  # the center of the node
        self._width = STANDARD_NODE_WIDTH
        self._height = STANDARD_NODE_HEIGHT
        self._half_width =  self._width / 2
        self._half_height = self._height / 2
        self._show_arrowheads = True
            
    def draw(self, screen):
        raise Exception("Node.raise must be implemented at subclass level.")

    def hook(self, hook_pos):
        if hook_pos == Hook.LEFT:
            return self._pos[0] - self._half_width, self._pos[1]
        elif hook_pos == Hook.RIGHT:
            return self._pos[0] + self._half_width, self._pos[1]
        elif hook_pos == Hook.TOP:
            return self._pos[0], self._pos[1] - self._half_height
        elif hook_pos == Hook.BOTTOM:
            return self._pos[0], self._pos[1] + self._half_height

    def set_pos(self, pos):
        self._pos = pos
    
    def get_pos(self):
        return self._pos
    
    def get_id(self):
        return self._model_node.get_id()
            

class PlaceViz(Node):
    def __init__(self, model_node):
        super().__init__(model_node)
    
    def draw(self, screen):
        pygame.draw.circle(screen, TUE_LIGHTBLUE, (self._pos[0], self._pos[1]), self._half_height)
        pygame.draw.circle(screen, TUE_BLUE, (self._pos[0], self._pos[1]), self._half_height, LINE_WIDTH)    
        font = pygame.font.SysFont('Calibri', TEXT_SIZE)
        bold_font = pygame.font.SysFont('Calibri', TEXT_SIZE, bold=True)

        # draw label
        label = font.render(self._model_node.get_id(), True, TUE_BLUE)
        text_x_pos = self._pos[0] - int(label.get_width()/2)
        text_y_pos = self._pos[1] + self._half_height + LINE_WIDTH
        screen.blit(label, (text_x_pos, text_y_pos))

        # draw marking
        mstr = "["
        ti = 0
        for token in self._model_node.marking:
            mstr += str(token.value) + "@" + str(round(token.time, 2))
            if ti < len(self._model_node.marking) - 1:
                mstr += ", "
            ti += 1
        mstr += "]"
        label = bold_font.render(mstr, True, TUE_RED)
        text_x_pos = self._pos[0] - int(label.get_width()/2)
        text_y_pos = self._pos[1] + self._half_height + LINE_WIDTH + int(label.get_height())
        screen.blit(label, (text_x_pos, text_y_pos))        


class TransitionViz(Node):
    def __init__(self, model_node):
        super().__init__(model_node)
    
    def draw(self, screen):
        pygame.draw.rect(screen, TUE_LIGHTBLUE, pygame.Rect(self._pos[0]-self._half_width, self._pos[1]-self._half_height, self._width, self._height))
        pygame.draw.rect(screen, TUE_BLUE, pygame.Rect(self._pos[0]-self._half_width, self._pos[1]-self._half_height, self._width, self._height), LINE_WIDTH)
        font = pygame.font.SysFont('Calibri', TEXT_SIZE)

        # draw label
        label = font.render(self._model_node.get_id(), True, TUE_BLUE)
        text_x_pos = self._pos[0] - int(label.get_width()/2)
        text_y_pos = self._pos[1] + self._half_height + LINE_WIDTH
        screen.blit(label, (text_x_pos, text_y_pos))


class Button(pygame.sprite.Sprite):
    def __init__(self):
        super().__init__()
        self.image = pygame.Surface((31, 31), pygame.SRCALPHA, 32)
        self.image = self.image.convert_alpha()
        self.image.set_alpha(128)
        self.rect = self.image.get_rect()

    def set_pos(self, x, y):
        self.rect.x = x
        self.rect.y = y

    def draw(self, screen):
        screen.blit(self.image, self.rect)
    

class Visualisation:
    """
    A class for visualizing the provided simulation problem as a Petri net.

    Attributes:
    - sim_problem (SimProblem): the simulation problem to visualize
    - layout_file (str): the file path to the layout file (optional)

    Methods:
    - save_layout(self, filename): saves the layout to a file
    - show(self): shows the visualisation
    """
    def __init__(self, sim_problem, layout_file=None):
        pygame.init()
        pygame.font.init()
        pygame.display.set_caption('Petri Net Visualisation')
        assets.create_assets(assets.images, "assets")
        icon = pygame.image.load('./assets/logo.png')
        pygame.display.set_icon(icon)

        self.__running = False
        self._problem = sim_problem
        self._nodes = dict()
        self._edges = []
        self._selected_nodes = None        

        # Add visualizations for prototypes, places, and transitions,
        # but not for places and transitions that are part of prototypes.
        element_to_prototype = dict()  # mapping of prototype element ids to prototype ids
        viznodes_with_edges = []
        for prototype in self._problem.prototypes:
            if prototype.visualize:
                prototype_viznode = prototype.get_visualisation()
                self._nodes[prototype.get_id()] = prototype_viznode
                viznodes_with_edges.append(prototype_viznode)
                for event in prototype.events:
                    element_to_prototype[event.get_id()] = prototype.get_id()
                for place in prototype.places:
                    element_to_prototype[place.get_id()] = prototype.get_id()
        for var in self._problem.places:
            if var.visualize and var.get_id() not in element_to_prototype:
                self._nodes[var.get_id()] = var.get_visualisation()
        for event in self._problem.events:
            if event.visualize and event.get_id() not in element_to_prototype:
                event_viznode = event.get_visualisation()
                self._nodes[event.get_id()] = event_viznode
                viznodes_with_edges.append(event_viznode)                
        # Add visualization for edges.
        # If an edge is from or to a prototype element, it must be from or to the prototype itself.
        for viznode in viznodes_with_edges:
            for incoming in viznode._model_node.incoming:
                node_id = incoming.get_id()
                if node_id.endswith(".queue"):
                    node_id = node_id[:-len(".queue")]
                if node_id in element_to_prototype:
                    node_id = element_to_prototype[node_id]
                if node_id in self._nodes:
                    other_viznode = self._nodes[node_id]
                    self._edges.append(Edge(start=(other_viznode, Hook.RIGHT), end=(viznode, Hook.LEFT)))
            for outgoing in viznode._model_node.outgoing:
                node_id = outgoing.get_id()
                if node_id.endswith(".queue"):
                    node_id = node_id[:-len(".queue")]
                if node_id in element_to_prototype:
                    node_id = element_to_prototype[node_id]
                if node_id in self._nodes:
                    other_viznode = self._nodes[node_id]
                    self._edges.append(Edge(start=(viznode, Hook.RIGHT), end=(other_viznode, Hook.LEFT)))
        layout_loaded = False
        if layout_file is not None:
            try:
                self.__load_layout(layout_file)
                layout_loaded = True
            except FileNotFoundError as e:
                print("WARNING: could not load the layout because of the exception below.\nauto-layout will be used.\n", e)
        if not layout_loaded:
            self.__layout()        

        self.__screen = pygame.display.set_mode(self._size, pygame.RESIZABLE)
        self._buttons = self.__init_buttons()
    
    def __draw(self):
        self.__screen.fill(TUE_GREY)
        for shape in self._nodes.values():
            shape.draw(self.__screen)
        for shape in self._edges:
            shape.draw(self.__screen)
        self.__draw_buttons()
        pygame.display.flip()

    def action_step(self):
        self._problem.step()
    
    def __init_buttons(self):
        # No buttons for now.
        # btn_step = Button()        
        # pygame.draw.polygon(btn_step.image, TUE_RED, [(0, 0), (0, 30), (20, 15)])
        # pygame.draw.polygon(btn_step.image, TUE_RED, [(20, 0), (20, 30), (25, 30), (25, 0)])
        # btn_step.set_pos(self._size[0]-100, self._size[1]-50)
        # btn_step.action = self.action_step

        # return [btn_step]
        return []

    def __draw_buttons(self):
        for btn in self._buttons:
            btn.draw(self.__screen)

    def __click_button_at(self, pos):
        for btn in self._buttons:
            if btn.rect.collidepoint(pos):
                btn.action()
                return True
        return False

    def __layout(self):
        graph = igraph.Graph()
        graph.to_directed()
        for node in self._nodes.values():
            graph.add_vertex(node.get_id())
        for edge in self._edges:            
            graph.add_edge(edge.get_start_node().get_id(), edge.get_end_node().get_id())
        layout = graph.layout_sugiyama()
        layout.rotate(-90)
        layout.scale(NODE_SPACING)
        boundaries = layout.boundaries(border=STANDARD_NODE_WIDTH)
        layout.translate(-boundaries[0][0], -boundaries[0][1])
        canvas_size = layout.boundaries(border=STANDARD_NODE_WIDTH*2)[1]
        self._size = (min(MAX_SIZE[0], canvas_size[0]), min(MAX_SIZE[1], canvas_size[1]))
        i = 0
        for v in graph.vs:
            xy = layout[i]
            xy  = (round(xy[0]/GRID_SPACING)*GRID_SPACING, round(xy[1]/GRID_SPACING)*GRID_SPACING)
            self._nodes[v["name"]].set_pos(xy)
            i += 1

    def save_layout(self, filename):
            """
            Saves the current layout of the nodes to a file.
            This method can be called after the show method.

            :param filename (str): The name of the file to save the layout to.
            """
            with open(filename, "w") as f:
                f.write(f"{int(self._size[0])},{int(self._size[1])}\n")
                for node in self._nodes.values():
                    if "," in node.get_id() or "\n" in node.get_id():
                        raise Exception("Node " + node.get_id() + ": Saving the layout cannot work if the node id contains a comma or hard return.")
                    f.write(f"{node.get_id()},{node.get_pos()[0]},{node.get_pos()[1]}\n")
    
    def __load_layout(self, filename):
        with open(filename, "r") as f:
            self._size = tuple(map(int, f.readline().strip().split(",")))
            for line in f:
                id, x, y = line.strip().split(",")
                if id in self._nodes:
                    self._nodes[id].set_pos((int(x), int(y)))

    def __get_node_at(self, pos):
        for node in self._nodes.values():
            if node.get_pos()[0] - node._width/2 <= pos[0] <= node.get_pos()[0] + node._width/2 and node.get_pos()[1] - node._height/2 <= pos[1] <= node.get_pos()[1] + node._height/2:
                return node
        return None

    def __drag(self, snap=False):
        nodes = self._selected_nodes[0]
        org_pos = self._selected_nodes[1]
        new_pos = pygame.mouse.get_pos()
        x_delta = new_pos[0] - org_pos[0]
        y_delta = new_pos[1] - org_pos[1]
        for node in nodes:
            new_x = node.get_pos()[0] + x_delta
            new_y = node.get_pos()[1] + y_delta
            if snap:
                new_x = round(new_x/GRID_SPACING)*GRID_SPACING
                new_y = round(new_y/GRID_SPACING)*GRID_SPACING
            node.set_pos((new_x, new_y))
        self._selected_nodes = nodes, new_pos

    def __handle_event(self, event):
        if event.type == pygame.QUIT:
            self.__running = False
        elif event.type == pygame.MOUSEBUTTONDOWN and event.button == 1:
            button_clicked = self.__click_button_at(event.pos)
            if not button_clicked:
                node = self.__get_node_at(event.pos)
                if node is not None:
                    self._selected_nodes = [node], event.pos
                else:
                    self._selected_nodes = self._nodes.values(), event.pos
        elif event.type == pygame.MOUSEBUTTONUP and event.button == 1 and self._selected_nodes is not None:
            self.__drag(snap=True)
            self._selected_nodes = None
        elif event.type == pygame.MOUSEMOTION and self._selected_nodes is not None:
            self.__drag()
        elif event.type == pygame.VIDEORESIZE:
            self._size = event.size
        elif event.type == pygame.KEYDOWN:
            if event.key == pygame.K_SPACE:
                self._problem.step()

    def show(self):
            """
            Displays the Petri net visualisation in a window.
            The method will block further execution until the window is closed.

            The visualisation can be interacted with using the mouse and keyboard.
            The spacebar can be used to step through the Petri net problem.
            The mouse can be used to drag nodes around.
            """
            clock = pygame.time.Clock()
            
            self.__running = True
            while self.__running:
                for event in pygame.event.get():
                    self.__handle_event(event)
                try:
                    self.__draw()
                except Exception:
                    print("Error while drawing the visualisation.")
                    print(traceback.format_exc())
                    self.__running = False
                clock.tick(30)

            pygame.quit()



import inspect
import pygame
from simpn.simulator import SimToken, SimVar
import random
import simpn.prototypes as prototypes
import simpn.visualisation as vis

class QueueingGenerator(prototypes.Prototype):

    def __init__(self, model, incoming, outgoing, name, interarrival_time, behavior=None):
        """
        Generates a composition of SimVar and SimEvent that represents a queueing system generator.
        Adds it to the specified model. The generator generates new cases with the specified interarrival_time.
        Cases are places on the outgoing SimVar (which can be a queue). The cases will be a tuple (unique_number, case_data).
        Case_data will be generated according to the specified behavior, or unspecified if behavior==None.
        The arrival of a new case will be signalled as a 'start event', so it can be used for process mining.

        :param model: the SimProblem to which the generator must be added.
        :param incoming: parameter is only here for consistency, must be [].
        :param outgoing: a list with a single SimVar in which the generated cases will be placed.
        :param name: the name of the generator.
        :param interarrival_time: the interarrival time with which events are generated. Can be a numeric value or a function that produces a numeric value, such as a sampling function from a random distribution.
        :param behavior: an optional behavior describing how case_data is produced.
        """
        super().__init__(model, incoming, outgoing, name)

        if len(incoming) != 0:
            raise TypeError("Generator " + name + ": cannot have any incoming.")
        if len(outgoing) != 1:
            raise TypeError("Generator " + name + ": must have exactly one outgoing.")
        if not callable(interarrival_time) and not type(interarrival_time) is int and not type(interarrival_time) is float:
            raise TypeError("Generator " + name + ": must either have a value or a function as interarrival_time.")
        interarrival_time_f = interarrival_time
        if type(interarrival_time) is int or type(interarrival_time) is float:
            interarrival_time_f = lambda: interarrival_time

        invar_name = name + "_timer"
        invar = model.add_var(invar_name)
        self.add_var(invar)
        if behavior is None:
            result = model.add_event([invar], [invar, outgoing[0]], lambda a: [SimToken(name + str(int(a[len(name):]) + 1), delay=interarrival_time_f()), SimToken((a,))], name=name + "<start_event>")
            self.add_event(result)
        else:
            if not callable(behavior):
                raise TypeError("Generator " + name + ": the behavior must be a function. (Maybe you made it a function call, exclude the brackets.)")
            if len(inspect.signature(behavior).parameters) != 0:
                raise TypeError("Generator " + name + ": the behavior function must not have many parameters.")
            result = model.add_event([invar], [invar, outgoing[0]], lambda a: [SimToken(name + str(int(a[len(name):]) + 1), delay=interarrival_time_f()), SimToken((a, behavior()[0].value))], name=name + "<start_event>")
            self.add_event(result)
        invar.put(name + "0")

        model.add_prototype(self)

    class QueueingGeneratorViz(vis.Node):
        def __init__(self, model_node):
            super().__init__(model_node)
        
        def draw(self, screen):
            x, y = self._pos
            hw, hh = self._half_width, self._half_height
            pygame.draw.polygon(screen, vis.TUE_LIGHTBLUE, [(x-hw, y-hh), (x+hw, y), (x-hw, y+hh), (x-hw, y-hh)])
            pygame.draw.polygon(screen, vis.TUE_BLUE, [(x-hw, y-hh), (x+hw, y), (x-hw, y+hh), (x-hw, y-hh)], vis.LINE_WIDTH)
            font = pygame.font.SysFont('Calibri', vis.TEXT_SIZE)

            # draw label
            label = font.render(self._model_node.get_id(), True, vis.TUE_BLUE)
            text_x_pos = self._pos[0] - int(label.get_width()/2)
            text_y_pos = self._pos[1] + self._half_height + vis.LINE_WIDTH
            screen.blit(label, (text_x_pos, text_y_pos))

    def get_visualisation(self):
        return self.QueueingGeneratorViz(self)


class QueueingQueue(SimVar):
    def __init__(self, model, _id, priority=lambda token: token.time):
        """
        A SimVar that represents a queue in a queueing system.
        It is just a SimVar with a different visualisation.
        """
        super().__init__(_id, priority)

        model.add_prototype_var(self)

    class QueueingQueueViz(vis.Node):
        def __init__(self, model_node):
            super().__init__(model_node)
        
        def draw(self, screen):
            x, y = self._pos
            hw, hh = self._half_width, self._half_height
            pygame.draw.rect(screen, vis.TUE_LIGHTBLUE, pygame.Rect(x-hw, y-hh, self._width, self._height))
            pygame.draw.line(screen, vis.TUE_BLUE, (x-hw, y-hh), (x+hw, y-hh), vis.LINE_WIDTH)
            pygame.draw.line(screen, vis.TUE_BLUE, (x+hw, y-hh), (x+hw, y+hh), vis.LINE_WIDTH)
            pygame.draw.line(screen, vis.TUE_BLUE, (x+hw, y+hh), (x-hw, y+hh), vis.LINE_WIDTH)

            pygame.draw.line(screen, vis.TUE_BLUE, (x-int(hw/2), y-hw), (x-int(hw/2), y+hw), vis.LINE_WIDTH)
            pygame.draw.line(screen, vis.TUE_BLUE, (x, y-hw), (x, y+hw), vis.LINE_WIDTH)
            pygame.draw.line(screen, vis.TUE_BLUE, (x+int(hw/2), y-hw), (x+int(hw/2), y+hw), vis.LINE_WIDTH)

            font = pygame.font.SysFont('Calibri', vis.TEXT_SIZE)
            bold_font = pygame.font.SysFont('Calibri', vis.TEXT_SIZE, bold=True)
            
            # draw label
            label = font.render(self._model_node.get_id(), True, vis.TUE_BLUE)
            text_x_pos = self._pos[0] - int(label.get_width()/2)
            text_y_pos = self._pos[1] + self._half_height + vis.LINE_WIDTH
            screen.blit(label, (text_x_pos, text_y_pos))

            # draw marking
            mstr = "["
            ti = 0
            for token in self._model_node.marking:
                mstr += str(token.value) + "@" + str(round(token.time, 2))
                if ti < len(self._model_node.marking) - 1:
                    mstr += ", "
                ti += 1
            mstr += "]"
            label = bold_font.render(mstr, True, vis.TUE_RED)
            text_x_pos = self._pos[0] - int(label.get_width()/2)
            text_y_pos = self._pos[1] + self._half_height + vis.LINE_WIDTH + int(label.get_height())
            screen.blit(label, (text_x_pos, text_y_pos))        

    def get_visualisation(self):
        return self.QueueingQueueViz(self)


class QueueingServer(prototypes.Prototype):
    def __init__(self, model, incoming, outgoing, name, processing_time, behavior=None, c=1):
        """
        Generates a composition of SimVar and SimEvent that represents a queue server with c resources.
        Adds it to the specified model. The server must have one incoming (queue) SimVar and one outgoing SimVar.
        The behavior specifies how the server may change the case data.
        By default the behavior is None, which means that the case data is not changed.
        It also specifies the processing time of the task in the form of a SimToken delay.

        :param model: the SimProblem to which the task composition must be added.
        :param incoming: a list with one SimVar.
        :param outgoing: a list with one SimVar.
        :param name: the name of the server.
        :param processing_time: the processing time of the server, which can either be a deterministic number or a function that is invoked to sample the processing time.
        :param behavior: the behavior function, which takes one input parameters according to the incoming and produces a single outgoing, which is a tuple (case, resource)@delay.
        :param c: the number of resources of the server.
        """
        super().__init__(model, incoming, outgoing, name)

        if len(incoming) != 1:
            raise TypeError("Server " + name + ": must have one input parameter for the case taken from the queue.")
        if len(outgoing) != 1:
            raise TypeError("Server " + name + ": must have one output parameter.")
        if behavior is not None and not callable(behavior):
            raise TypeError("Server " + name + ": the behavior must be a function. (Maybe you made it a function call, exclude the brackets.)")
        if behavior is not None and len(inspect.signature(behavior).parameters) != 1:
            raise TypeError("Server " + name + ": the behavior function must have one parameter for the case taken from the queue.")
        if not callable(processing_time) and not type(processing_time) is int and not type(processing_time) is float:
            raise TypeError("Server " + name + ": must either have a value or a function as processing_time.")
        processing_time_f = processing_time
        if type(processing_time) is int or type(processing_time) is float:
            processing_time_f = lambda: processing_time

        resource_name = name + "_resource"
        busyvar_name = name + "_busy"
        start_event_name = name + "<task:start>"
        complete_event_name = name + "<task:complete>"
        self._busyvar = model.add_var(busyvar_name)
        self.add_var(self._busyvar)
        self._resourcevar = model.add_var(resource_name)
        self.add_var(self._resourcevar)
        for i in range(c):
            self._resourcevar.put("r"+str(i))
        if behavior is None:
            behavior_f = lambda a, r: [SimToken((a, r), delay=processing_time_f())]
        else:
            behavior_f = lambda a, r: [SimToken((behavior(a), r), delay=processing_time_f())]
        start_event = model.add_event([incoming[0], self._resourcevar], [self._busyvar], behavior_f, name=start_event_name)
        self.add_event(start_event)
        complete_event = model.add_event([self._busyvar], [outgoing[0], self._resourcevar], lambda b: [SimToken(b[0]), SimToken(b[1])], name=complete_event_name)
        self.add_event(complete_event)

        model.add_prototype(self)

    class QueueingServerViz(vis.Node):
        def __init__(self, model_node):
            super().__init__(model_node)
        
        def draw(self, screen):
            pygame.draw.circle(screen, vis.TUE_LIGHTBLUE, (self._pos[0], self._pos[1]), self._half_height)
            pygame.draw.circle(screen, vis.TUE_BLUE, (self._pos[0], self._pos[1]), self._half_height, vis.LINE_WIDTH)    
            font = pygame.font.SysFont('Calibri', vis.TEXT_SIZE)
            bold_font = pygame.font.SysFont('Calibri', vis.TEXT_SIZE, bold=True)

            # draw label
            label = font.render(self._model_node.get_id(), True, vis.TUE_BLUE)
            text_x_pos = self._pos[0] - int(label.get_width()/2)
            text_y_pos = self._pos[1] + self._half_height + vis.LINE_WIDTH
            screen.blit(label, (text_x_pos, text_y_pos))

            # draw marking
            mstr = "["
            ti = 0
            for token in self._model_node._busyvar.marking:
                mstr += str(token.value) + "@" + str(round(token.time, 2))
                if ti < len(self._model_node._busyvar.marking) - 1:
                    mstr += ", "
                ti += 1
            mstr += "]"
            label = bold_font.render(mstr, True, vis.TUE_RED)
            text_x_pos = self._pos[0] - int(label.get_width()/2)
            text_y_pos = self._pos[1] + self._half_height + vis.LINE_WIDTH + int(label.get_height())
            screen.blit(label, (text_x_pos, text_y_pos))        

            # draw free servers
            label = bold_font.render("free: " + str(len(self._model_node._resourcevar.marking)), True, vis.TUE_RED)
            text_x_pos = self._pos[0] - int(label.get_width()/2)
            text_y_pos = self._pos[1] - self._half_height - vis.LINE_WIDTH - int(label.get_height())
            screen.blit(label, (text_x_pos, text_y_pos))        

    def get_visualisation(self):
        return self.QueueingServerViz(self)


class QueueingSink(SimVar):
    def __init__(self, model, _id, priority=lambda token: token.time):
        """
        A SimVar that represents a sink.
        It is just a SimVar with a different visualisation.
        """
        super().__init__(_id, priority)

        model.add_prototype_var(self)

    class QueueingSinkViz(vis.Node):
        def __init__(self, model_node):
            super().__init__(model_node)
        
        def draw(self, screen):
            x, y = self._pos
            hw, hh = self._half_width, self._half_height
            pygame.draw.polygon(screen, vis.TUE_LIGHTBLUE, [(x-hw, y), (x+hw, y-hw), (x+hw, y+hh), (x-hw, y)])
            pygame.draw.polygon(screen, vis.TUE_BLUE, [(x-hw, y), (x+hw, y-hw), (x+hw, y+hh), (x-hw, y)], vis.LINE_WIDTH)

            font = pygame.font.SysFont('Calibri', vis.TEXT_SIZE)
            bold_font = pygame.font.SysFont('Calibri', vis.TEXT_SIZE, bold=True)
            
            # draw label
            label = font.render(self._model_node.get_id(), True, vis.TUE_BLUE)
            text_x_pos = self._pos[0] - int(label.get_width()/2)
            text_y_pos = self._pos[1] + self._half_height + vis.LINE_WIDTH
            screen.blit(label, (text_x_pos, text_y_pos))

            # draw marking
            label = bold_font.render(str(len(self._model_node.marking)), True, vis.TUE_RED)
            text_x_pos = self._pos[0] - int(label.get_width()/2)
            text_y_pos = self._pos[1] + self._half_height + vis.LINE_WIDTH + int(label.get_height())
            screen.blit(label, (text_x_pos, text_y_pos))        

    def get_visualisation(self):
        return self.QueueingSinkViz(self)


class QueueingChoice(prototypes.Prototype):

    def __init__(self, model, incoming, outgoing, name, weights):
        """
        Generates a composition of SimVar and SimEvent that represents a choice of which queue to go in.
        Adds it to the specified model. The choice is made according to the weights.
        Cases are taken from the (single) incoming SimVar and placed in one of the outgoing SimVars according to specified the weights.

        :param model: the SimProblem to which the generator must be added.
        :param incoming: parameter is only here for consistency, must be [].
        :param outgoing: a list with a single SimVar in which the generated cases will be placed.
        :param name: the name of the generator.
        :param weights: a list of weights that sum up to 1.0.
        """
        super().__init__(model, incoming, outgoing, name)

        if len(incoming) != 1:
            raise TypeError("Choice " + name + ": must have exacly one incoming SimVar.")
        if len(outgoing) < 2:
            raise TypeError("Choice " + name + ": must have at least two outgoing SimVar.")
        if len(outgoing) != len(weights):
            raise TypeError("Choice " + name + ": the number of outgoing SimVars must be equal to the number of weights.")
        for w in weights:
            if type(w) is not int and type(w) is not float:
                raise TypeError("Choice " + name + ": the weights must be numeric values. " + str(w) + " is not.")

        def choose(c):
            chosen = random.choices(list(range(len(outgoing))), weights)[0]
            result = [None] * len(outgoing)
            result[chosen] = SimToken(c)
            return result
        choice_event = model.add_event(incoming, outgoing, choose)
        self.add_event(choice_event)

        model.add_prototype(self)

    class QueueingChoiceViz(vis.Node):
        def __init__(self, model_node):
            super().__init__(model_node)
        
        def draw(self, screen):
            x, y = self._pos
            hw, hh = self._half_width, self._half_height
            pygame.draw.polygon(screen, vis.TUE_LIGHTBLUE, [(x-hw, y), (x, y-hw), (x+hw, y), (x, y+hh), (x-hw, y)])
            pygame.draw.polygon(screen, vis.TUE_BLUE, [(x-hw, y), (x, y-hw), (x+hw, y), (x, y+hh), (x-hw, y)], vis.LINE_WIDTH)
            font = pygame.font.SysFont('Calibri', vis.TEXT_SIZE)

            # draw label
            label = font.render(self._model_node.get_id(), True, vis.TUE_BLUE)
            text_x_pos = self._pos[0] - int(label.get_width()/2)
            text_y_pos = self._pos[1] + self._half_height + vis.LINE_WIDTH
            screen.blit(label, (text_x_pos, text_y_pos))

    def get_visualisation(self):
        return self.QueueingChoiceViz(self)



from datetime import datetime, timedelta
from enum import Enum, auto


class TimeUnit(Enum):
    """An enumeration for the unit in which simulation time is measured."""
    SECONDS = auto()
    """
    Measured in seconds.
    
    :meta hide-value:
    """
    MINUTES = auto()
    """
    Measured in minutes.

    :meta hide-value:
    """
    HOURS = auto()
    """
    Measured in hours.

    :meta hide-value:
    """
    DAYS = auto()
    """
    Measured in days.

    :meta hide-value:
    """


class Reporter:
    """
    A reporter can be passed to the simpn.simulator.SimProblem.simulate function to report on what happens during the simulator.
    To this end, a reported must implement the callback function.
    """
    def callback(self, timed_binding):
        """
        A function that is invoked by a simpn.simulator.SimProblem each time a event happened.
        It receives a timed_binding, which is a triple (binding, time, event):
        - binding = [(v1: SimVar, t1: SimToken), (v2: SimVar, t2: SimToken), ...] of the variable values that caused the event.
        - time is the simulation time at which the event happened.
        - event: SimEvent is the event that happened.

        :param timed_binding: the triple (binding, time, event) that described the event that happened with its variable values and the time at which it happened.
        """
        raise NotImplementedError


class SimpleReporter(Reporter):
    """
    A simple reporter that just prints the occurring events to the standard output.
    """
    def callback(self, timed_binding):
        result = str(timed_binding[2]) + "{"
        i = 0
        for var, token in timed_binding[0]:
            result += str(var) + ": " + str(token.value)
            if i < len(timed_binding[0])-1:
                result += ", "
            i += 1
        result += "}" + "@t=" + str(timed_binding[1])
        print(result)


class WarmupReporter(Reporter):
    """
    A reporter that reports on the warmup period of a simulation by computing the average cycle times over time.
    It assumes tasks are executed for cases that arrive via start_events and complete at end_events.
    """
    def __init__(self):
        self.times = [0]
        self.average_cycle_times = [0]

        self.__total_times = [0]
        self.__nr_completed_cases = 0
        self.__case_arrivals = {}

    def callback(self, timed_binding):
        (binding, time, event) = timed_binding
        if event.get_id().endswith("<start_event>"):
            case_id = binding[0][1].value  # the case_id is always [0] the first variable in the binding, the [1] token value of that, and [0] the case_id of the value.
            self.__case_arrivals[case_id] = time
        elif event.get_id().endswith("<end_event>"):
            case_id = binding[0][1].value[0]  # the case_id is always [0] the first variable in the binding, the [1] token value of that, and [0] the case_id of the value.
            waiting_time = time - self.__case_arrivals[case_id]
            total_time = self.__total_times[self.__nr_completed_cases]
            total_time += waiting_time
            self.__nr_completed_cases += 1
            self.times.append(time)
            self.__total_times.append(total_time)
            self.average_cycle_times.append(total_time/self.__nr_completed_cases)


class ProcessReporter(Reporter):
    """
    A reporter that heavily depends on the process prototypes (task, start_event, intermediate_event, end_event) to report on what happens.
    It assumes tasks are executed for cases that arrive via start_events and complete at end_events. It measures:
    - nr_started: the number of cases that started.
    - nr_completed: the number of cases that completed.
    - total_wait_time: the sum of waiting times of completed cases.
    - total_proc_time: the sum of processing times of completed cases.
    - total_cycle_time: the sum of cycle times of completed cases.
    - resource_busy_times: a mapping of resource_id -> the time the resource was busy during simulation.
    """

    def __init__(self, warmup_time=0):
        self.resource_busy_times = dict()  # mapping of resource_id -> the time the resource was busy during simulation.
        self.nr_started = 0  # number of cases that started
        self.nr_completed = 0  # number of cases that completed
        self.total_wait_time = 0  # sum of waiting times of completed cases
        self.total_proc_time = 0  # sum of processing times of completed cases
        self.total_cycle_time = 0  # sum of cycle times of completed cases

        self.__status = dict()  # case_id -> (nr_busy_tasks, arrival_time, sum_wait_times, sum_proc_times, time_last_busy_change); time_last_busy_change is the time nr_busy_tasks last went from 0 to 1 or from 1 to 0
        self.__resource_start_times = dict()  # resource -> time
        self.__last_time = 0

        self.warmup_time = warmup_time

    def callback(self, timed_binding):
        (binding, time, event) = timed_binding
        self.__last_time = time
        if event.get_id().endswith("<start_event>"):
            case_id = binding[0][1].value  # the case_id is always [0] the first variable in the binding, the [1] token value of that, and [0] the case_id of the value.
            self.__status[case_id] = (0, time, 0, 0, time)
            if time > self.warmup_time:
                self.nr_started += 1
        elif event.get_id().endswith("<task:start>"):
            case_id = binding[0][1].value[0]  # the case_id is always [0] the first variable in the binding, the [1] token value of that, and [0] the case_id of the value.
            resource_id = binding[1][1].value  # the resource_id is always [1] the second variable in the binding, the [1] token value of that.
            self.__resource_start_times[resource_id] = time
            (nr_busy_tasks, arrival_time, sum_wait_times, sum_proc_times, time_last_busy_change) = self.__status[case_id]
            if nr_busy_tasks == 0:
                sum_wait_times += time - time_last_busy_change
                time_last_busy_change = time
            nr_busy_tasks += 1
            self.__status[case_id] = (nr_busy_tasks, arrival_time, sum_wait_times, sum_proc_times, time_last_busy_change)
        elif event.get_id().endswith("<task:complete>"):
            case_id = binding[0][1].value[0][0]  # the case_id is always [0] the first variable in the binding, the [1] token value of that, and [0] the case_id of the value.
            resource_id = binding[0][1].value[1]  # the resource_id is always [0] the first variable in the binding, the [1] token value of that, and [1] the resource_id of the value.
            if time > self.warmup_time:
                if resource_id not in self.resource_busy_times.keys():
                    self.resource_busy_times[resource_id] = 0
                if self.__resource_start_times[resource_id] < self.warmup_time:
                    self.resource_busy_times[resource_id] += time - self.warmup_time
                else:
                    self.resource_busy_times[resource_id] += time - self.__resource_start_times[resource_id]
            del self.__resource_start_times[resource_id]
            (nr_busy_tasks, arrival_time, sum_wait_times, sum_proc_times, time_last_busy_change) = self.__status[case_id]
            if nr_busy_tasks == 1:
                sum_proc_times += time - time_last_busy_change
                time_last_busy_change = time
            nr_busy_tasks -= 1
            self.__status[case_id] = (nr_busy_tasks, arrival_time, sum_wait_times, sum_proc_times, time_last_busy_change)
        elif event.get_id().endswith("<end_event>"):
            case_id = binding[0][1].value[0]  # the case_id is always [0] the first variable in the binding, the [1] token value of that, and [0] the case_id of the value.
            (nr_busy_tasks, arrival_time, sum_wait_times, sum_proc_times, time_last_busy_change) = self.__status[case_id]
            del self.__status[case_id]
            if arrival_time > self.warmup_time:
                self.nr_completed += 1
                self.total_wait_time += sum_wait_times
                self.total_proc_time += sum_proc_times
                self.total_cycle_time += time - arrival_time

    def print_result(self):
        print("Nr. cases started:            ", self.nr_started)
        print("Nr. cases completed:          ", self.nr_completed)
        print("Avg. waiting time per case:   ", round(self.total_wait_time/self.nr_completed, 3))
        print("Avg. processing time per case:", round(self.total_proc_time/self.nr_completed, 3))
        print("Avg. cycle time per case:     ", round(self.total_cycle_time/self.nr_completed, 3))

        # process the resources that are currently busy
        for resource_id in self.__resource_start_times.keys():
            self.resource_busy_times[resource_id] += self.__last_time - self.__resource_start_times[resource_id]

        self.__resource_start_times.clear()
        
        for resource_id in self.resource_busy_times.keys():
            print("Resource", resource_id, "utilization:", round(self.resource_busy_times[resource_id]/(self.__last_time - self.warmup_time), 2))

class EventLogReporter(Reporter):
    """
    A reporter that heavily depends on the process prototypes (task, start_event, intermediate_event, end_event) to report on what happens.
    It stores the events that happen in an event log that can be mined with some process mining tool.
    As simulation time is a numerical value, some processing is done to convert simulation time
    into a time format that can be read and interpreted by a process mining tool. To that end,
    the timeunit in which simulation time is measured must be passed as well as the
    initial_time moment in calendar time from which the simulation time will be measured.
    A particular simulation_time moment will then be stored in the log as:
    initial_time + simulation_time timeunits. Data can also be reported on by specifying the
    corresponding data fields. The log will contain these data fields as additional columns in the order specified.
    The case that is processed in each task is assumed to have the same number of data fields as the data_fields parameter.
    The data of each case must initially be generated by the start event and can be modified by tasks.
    Each task must pass the same data fields, optionally with modified values.

    :param filename: the name of the file in which the event log must be stored.
    :param timeunit: the :class:`.TimeUnit` of simulation time.
    :param initial_time: a datetime value.
    :param time_format: a datetime formatting string.
    :param data_fields: the data fields to report in the log.
    :param separator: the separator to use in the log.
    :param data_fields: a list of data field names to include in the log.
    """
    def __init__(self, filename, timeunit=TimeUnit.MINUTES, initial_time=datetime(2020, 1, 1), time_format="%Y-%m-%d %H:%M:%S.%f", separator=",", data_fields=None):
        self.task_start_times = dict()
        self.timeunit = timeunit
        self.initial_time = initial_time
        self.time_format = time_format
        self.logfile = open(filename, "wt")
        self.sep = separator        
        self.data_fields = data_fields
        self.logfile.write("case_id"+self.sep+"task"+self.sep+"resource"+self.sep+"start_time"+self.sep+"completion_time")
        if data_fields is not None:
            for field in data_fields:
                self.logfile.write(self.sep + field)
        self.logfile.write("\n")

    def displace(self, time):
        return self.initial_time + (timedelta(seconds=time) if self.timeunit == TimeUnit.SECONDS else timedelta(
            minutes=time) if self.timeunit == TimeUnit.MINUTES else timedelta(
            hours=time) if self.timeunit == TimeUnit.HOURS else timedelta(
            days=time) if self.timeunit == TimeUnit.DAYS else None)

    def callback(self, timed_binding):
        (binding, time, event) = timed_binding        
        if event.get_id().endswith("<task:start>"):            
            case_id = binding[0][1].value[0]
            task = event.get_id()[:event.get_id().index("<")]
            self.task_start_times[(case_id, task)] = time
        elif event.get_id().endswith("<task:complete>"):
            case_id = binding[0][1].value[0][0]
            task = event.get_id()[:event.get_id().index("<")]
            resource = binding[0][1].value[1]
            if (case_id, task) in self.task_start_times.keys():
                self.logfile.write(str(case_id) + self.sep)
                self.logfile.write(task + self.sep)
                self.logfile.write(str(resource) + self.sep)
                self.logfile.write(self.displace(self.task_start_times[(case_id, task)]).strftime(self.time_format) + self.sep)
                self.logfile.write(self.displace(time).strftime(self.time_format))
                if self.data_fields is not None:
                    data = binding[0][1].value[0][1:]
                    data_i = 0
                    for _ in self.data_fields:
                        self.logfile.write(self.sep)
                        if data_i < len(data):
                            self.logfile.write(str(data[data_i]))
                        data_i += 1
                self.logfile.write("\n")
                self.logfile.flush()
                del self.task_start_times[(case_id, task)]
        elif event.get_id().endswith("<start_event>") or event.get_id().endswith("<intermediate_event>") or event.get_id().endswith("<end_event>"):
            if event.get_id().endswith("<start_event>"):
                case_id = binding[0][1].value
            else:
                case_id = binding[0][1].value[0]
            event = event.get_id()[:event.get_id().index("<")]
            self.logfile.write(str(case_id) + self.sep)
            self.logfile.write(event + self.sep)
            self.logfile.write(self.sep)
            self.logfile.write(self.displace(time).strftime(self.time_format) + self.sep)
            self.logfile.write(self.displace(time).strftime(self.time_format))
            if self.data_fields is not None:
                for _ in self.data_fields:
                    self.logfile.write(self.sep)
            self.logfile.write("\n")
            self.logfile.flush()

    def close(self):
        self.logfile.close()


import inspect
import pygame
from simpn.simulator import SimToken, SimVar, SimEvent
import simpn.visualisation as vis


class Prototype:
    """
    Superclass for all prototypes. Contains the basic structure of a prototype, which is a composition of SimVar and SimEvent.
    A prototype must subclass this class.
    Each event and variable that the prototype creates must both be added to the model and to the prototype itself.
    The prototype must finally be added to the model, using the model.add_prototype() method.
    """

    def __init__(self, model, incoming, outgoing, name):
        self.model = model
        self.incoming = incoming
        self.outgoing = outgoing
        self.name = name
        self.places = []
        self.events = []
        self.visualize = True

    def add_var(self, var):
        self.places.append(var)

    def add_event(self, event):
        self.events.append(event)

    def get_id(self):
        return self.name
    
    def get_visualisation(self):
        raise NotImplementedError("Method get_visualisation must be implemented in subclass.")

    def set_invisible(self):
        self.visualize = False

class BPMNStartEvent(Prototype):  
    """
    Generates a composition of SimVar and SimEvent that represents a BPMN start event.
    Adds it to the specified model. The start event generates new cases with the specified interarrival_time.
    Cases are places on the outgoing SimVar. The cases will be a tuple (unique_number, case_data).
    Case_data will be generated according to the specified behavior, or unspecified if behavior==None.

    :param model: the SimProblem to which the start event composition must be added.
    :param incoming: parameter is only here for consistency, must be [].
    :param outgoing: a list with a single SimVar in which the cases will be placed.
    :param name: the name of the start event.
    :param interarrival_time: the interarrival time with which events are generated. Can be a numeric value or a function that produces a numeric value, such as a sampling function from a random distribution.
    :param behavior: an optional behavior describing how case_data is produced.
    """

    def __init__(self, model, incoming, outgoing, name, interarrival_time, behavior=None):
        super().__init__(model, incoming, outgoing, name)

        if len(incoming) != 0:
            raise TypeError("Start event " + name + ": cannot have any incoming.")
        if len(outgoing) != 1:
            raise TypeError("Start event " + name + ": must have exactly one outgoing.")
        if not callable(interarrival_time) and not type(interarrival_time) is int and not type(interarrival_time) is float:
            raise TypeError("Start event " + name + ": must either have a value or a function as interarrival_time.")
        interarrival_time_f = interarrival_time
        if type(interarrival_time) is int or type(interarrival_time) is float:
            interarrival_time_f = lambda: interarrival_time

        invar_name = name + "_timer"
        invar = model.add_var(invar_name)
        self.add_var(invar)
        if behavior is None:
            result = model.add_event([invar], [invar, outgoing[0]], lambda a: [SimToken(name + str(int(a[len(name):]) + 1), delay=interarrival_time_f()), SimToken((a,))], name=name + "<start_event>")
            self.add_event(result)
        else:
            if not callable(behavior):
                raise TypeError("Start event " + name + ": the behavior must be a function. (Maybe you made it a function call, exclude the brackets.)")
            if len(inspect.signature(behavior).parameters) != 0:
                raise TypeError("Start event " + name + ": the behavior function must not have many parameters.")
            result = model.add_event([invar], [invar, outgoing[0]], lambda a: [SimToken(name + str(int(a[len(name):]) + 1), delay=interarrival_time_f()), SimToken((a, behavior()[0].value))], name=name + "<start_event>")
            self.add_event(result)
        invar.put(name + "0")

        model.add_prototype(self)

    class BPMNStartEventViz(vis.Node):
        def __init__(self, model_node):
            super().__init__(model_node)
        
        def draw(self, screen):
            pygame.draw.circle(screen, vis.TUE_LIGHTBLUE, (self._pos[0], self._pos[1]), self._width/2)
            pygame.draw.circle(screen, vis.TUE_BLUE, (self._pos[0], self._pos[1]), self._width/2, vis.LINE_WIDTH)    
            font = pygame.font.SysFont('Calibri', vis.TEXT_SIZE)

            # draw label
            label = font.render(self._model_node.get_id(), True, vis.TUE_BLUE)
            text_x_pos = self._pos[0] - int(label.get_width()/2)
            text_y_pos = self._pos[1] + self._half_height + vis.LINE_WIDTH
            screen.blit(label, (text_x_pos, text_y_pos))

    def get_visualisation(self):
        return self.BPMNStartEventViz(self)
    

class BPMNTask(Prototype):
    """
    Generates a composition of SimVar and SimEvent that represents a BPMN task.
    Adds it to the specified model. The task must have at least two incoming and two outgoing SimVar.
    The first SimVar represents the case that must be processed by the task and the second the resource.
    There can be additional SimVar that represent additional variables that the task may need or produce.
    The behavior specifies how the task may change the case data.
    It also specifies the processing time of the task in the form of a SimToken delay.
    The behavior must take input parameters according to the incoming variables and produces a single intermediate variable, which must be a tuple (case, resource [, <additional variable 1>, <additional variable 2>, ...])@delay.

    :param model: the SimProblem to which the task composition must be added.
    :param incoming: a list with at least two SimVar: a case SimVar, a resource SimVar, and optionally additional SimVar.
    :param outgoing: a list with at least two SimVar: a case SimVar, a resource SimVar, and optionally additional SimVar.
    :param name: the name of the task.
    :param behavior: the behavior function, which takes at least two input parameters according to the incoming variables and produces a single intermediate variable.
    :param guard: an optional guard that specifies which combination of case and resource is allowed. The guard must take at least two input parameters according to the incoming variables.
    :param outgoing_behavior: an optional behavior that specifies how the outgoing variables are produced from the single intermediate variable.
    """

    def __init__(self, model, incoming, outgoing, name, behavior, guard=None, outgoing_behavior=None):
        super().__init__(model, incoming, outgoing, name)

        if len(incoming) < 2:
            raise TypeError("Task event " + name + ": must have at least two input parameters; the first for cases and the second for resources.")
        if len(outgoing) < 2:
            raise TypeError("Task event " + name + ": must have at least two output parameters; the first for cases and the second for resources.")
        if not callable(behavior):
            raise TypeError("Task event " + name + ": the behavior must be a function. (Maybe you made it a function call, exclude the brackets.)")
        if len(inspect.signature(behavior).parameters) != len(incoming):
            raise TypeError("Task event " + name + ": the behavior function must have as many parameters as the number of incoming variables.")
        if outgoing_behavior is not None:
            if not callable(outgoing_behavior):
                raise TypeError("Task event " + name + ": the outgoing_behavior must be a function. (Maybe you made it a function call, exclude the brackets.)")
            if len(inspect.signature(outgoing_behavior).parameters) != 1:
                raise TypeError("Task event " + name + ": the outgoing_behavior function must have exactly one parameter.")
        if guard is not None:
            if not callable(guard):
                raise TypeError("Task event " + name + ": the guard must be a function. (Maybe you made it a function call, exclude the brackets.)")
            if len(inspect.signature(guard).parameters) != len(incoming):
                raise TypeError("Task event " + name + ": the guard function must have as many parameters as the number of incoming variables.")

        busyvar_name = name + "_busy"
        start_event_name = name + "<task:start>"
        complete_event_name = name + "<task:complete>"
        self._busyvar = model.add_var(busyvar_name)
        self.add_var(self._busyvar)
        start_event = model.add_event(incoming, [self._busyvar], behavior, name=start_event_name, guard=guard)
        self.add_event(start_event)
        if outgoing_behavior is None:
            complete_event = model.add_event([self._busyvar], outgoing, lambda b: [SimToken(b[i]) for i in range(len(b))], name=complete_event_name)
        else:
            complete_event = model.add_event([self._busyvar], outgoing, outgoing_behavior, name=complete_event_name)
        self.add_event(complete_event)

        model.add_prototype(self)

    class BPMNTaskViz(vis.Node):
        def __init__(self, model_node):
            super().__init__(model_node)
            self._width = 100
            self._height = vis.STANDARD_NODE_HEIGHT
            self._half_width =  self._width / 2
            self._half_height = self._height / 2

        def draw(self, screen):
            x_pos, y_pos = int(self._pos[0] - self._width/2), int(self._pos[1] - self._height/2)
            pygame.draw.rect(screen, vis.TUE_LIGHTBLUE, pygame.Rect(x_pos, y_pos, self._width, self._height), border_radius=int(0.075*self._width))
            pygame.draw.rect(screen, vis.TUE_BLUE, pygame.Rect(x_pos, y_pos, self._width, self._height),  vis.LINE_WIDTH, int(0.075*self._width))
            font = pygame.font.SysFont('Calibri', vis.TEXT_SIZE)
            bold_font = pygame.font.SysFont('Calibri', vis.TEXT_SIZE, bold=True)

            # draw label
            label = font.render(self._model_node.get_id(), True, vis.TUE_BLUE)
            text_x_pos = int((self._width - label.get_width())/2) + x_pos
            text_y_pos = int((self._height - label.get_height())/2) + y_pos
            screen.blit(label, (text_x_pos, text_y_pos))

            # draw marking
            mstr = "["
            ti = 0
            for token in self._model_node._busyvar.marking:
                mstr += str(token.value) + "@" + str(round(token.time, 2))
                if ti < len(self._model_node._busyvar.marking) - 1:
                    mstr += ", "
                ti += 1
            mstr += "]"
            label = bold_font.render(mstr, True, vis.TUE_RED)
            text_x_pos = self._pos[0] - int(label.get_width()/2)
            text_y_pos = self._pos[1] + self._half_height + vis.LINE_WIDTH
            screen.blit(label, (text_x_pos, text_y_pos))        


    def get_visualisation(self):
        return self.BPMNTaskViz(self)


class BPMNIntermediateEvent(Prototype):
    """
    Generates a composition of SimVar and SimEvent that represents a BPMN intermediate event.
    The intermediate event can make changes to the data of a case and can generate waiting time for the case.

    :param model: the SimProblem to which the event composition must be added.
    :param incoming: a list with at least one SimVar: a case SimVar.
    :param outgoing: a list with at least one SimVar: a case SimVar.
    :param name: the name of the event.
    :param behavior: specifies the changes that the intermediate event makes to the data and the delay that the intermediate event may lead to.
    :param guard: an optional guard that specifies under which condition the intermediate event can happen.
    """

    def __init__(self, model, incoming, outgoing, name, behavior, guard=None):
        super().__init__(model, incoming, outgoing, name)

        if len(incoming) < 1:
            raise TypeError("Event " + name + ": must have at least one input parameter for cases.")
        if len(outgoing) < 1:
            raise TypeError("Event " + name + ": must have at least one output parameter for cases.")
        if not callable(behavior):
            raise TypeError("Event " + name + ": the behavior must be a function. (Maybe you made it a function call, exclude the brackets.)")
        if len(inspect.signature(behavior).parameters) != len(incoming):
            raise TypeError("Event " + name + ": the behavior function must have as many parameters as the number of incoming variables.")
        if guard is not None:
            if not callable(guard):
                raise TypeError("Event " + name + ": the guard must be a function. (Maybe you made it a function call, exclude the brackets.)")
            if len(inspect.signature(guard).parameters) != len(incoming):
                raise TypeError("Event " + name + ": the guard function must have as many parameters as the number of incoming variables.")

        result = model.add_event(incoming, outgoing, behavior, name=name + "<intermediate_event>", guard=guard)
        self.add_event(result)

        model.add_prototype(self)

    class BPMNIntermediateEventViz(vis.Node):
        def __init__(self, model_node):
            super().__init__(model_node)
        
        def draw(self, screen):
            pygame.draw.circle(screen, vis.TUE_LIGHTBLUE, (self._pos[0], self._pos[1]), self._width/2)
            pygame.draw.circle(screen, vis.TUE_BLUE, (self._pos[0], self._pos[1]), self._width/2, vis.LINE_WIDTH)   
            pygame.draw.circle(screen, vis.TUE_BLUE, (self._pos[0], self._pos[1]), self._width/2-3, vis.LINE_WIDTH)   
            font = pygame.font.SysFont('Calibri', vis.TEXT_SIZE)

            # draw label
            label = font.render(self._model_node.get_id(), True, vis.TUE_BLUE)
            text_x_pos = self._pos[0] - int(label.get_width()/2)
            text_y_pos = self._pos[1] + self._half_height + vis.LINE_WIDTH
            screen.blit(label, (text_x_pos, text_y_pos))

    def get_visualisation(self):
        return self.BPMNIntermediateEventViz(self)


class BPMNEndEvent(Prototype):
    """
    Generates a composition of SimVar and SimEvent that represents a BPMN end event.

    :param model: the SimProblem to which the event composition must be added.
    :param incoming: a list with one SimVar: a case SimVar.
    :param outgoing: parameter is only here for consistency, must be [].
    :param name: the name of the event.
    """

    def __init__(self, model, incoming, outgoing, name):
        super().__init__(model, incoming, outgoing, name)

        if len(incoming) != 1:
            raise TypeError("Event " + name + ": must have one input parameter for cases.")
        if len(outgoing) != 0:
            raise TypeError("Event " + name + ": must not have output parameters.")

        result = model.add_event(incoming, outgoing, lambda c: [], name=name + "<end_event>")
        self.add_event(result)

        model.add_prototype(self)

    class BPMNEndEventViz(vis.Node):
        def __init__(self, model_node):
            super().__init__(model_node)
        
        def draw(self, screen):
            pygame.draw.circle(screen, vis.TUE_LIGHTBLUE, (self._pos[0], self._pos[1]), self._width/2)
            pygame.draw.circle(screen, vis.TUE_BLUE, (self._pos[0], self._pos[1]), self._width/2, vis.LINE_WIDTH*2)
            font = pygame.font.SysFont('Calibri', vis.TEXT_SIZE)

            # draw label
            label = font.render(self._model_node.get_id(), True, vis.TUE_BLUE)
            text_x_pos = self._pos[0] - int(label.get_width()/2)
            text_y_pos = self._pos[1] + self._half_height + vis.LINE_WIDTH
            screen.blit(label, (text_x_pos, text_y_pos))

    def get_visualisation(self):
        return self.BPMNEndEventViz(self)


class BPMNFlow(SimVar):
    """
    A SimVar that represents a BPMN Flow.
    It is just a SimVar with a different visualisation.
    """

    def __init__(self, model, _id):
        super().__init__(_id)

        model.add_prototype_var(self)

    class BPMNFlowViz(vis.Node):
        def __init__(self, model_node):
            super().__init__(model_node)
            self._width = 3
            self._height = 3
            self._half_width =  self._width / 2
            self._half_height = self._height / 2
            self._show_arrowheads = False
        
        def draw(self, screen):
            x, y = self._pos
            pygame.draw.circle(screen, vis.TUE_BLUE, (x, y), self._width)

            bold_font = pygame.font.SysFont('Calibri', vis.TEXT_SIZE, bold=True)
            
            # draw marking
            mstr = "["
            ti = 0
            for token in self._model_node.marking:
                mstr += str(token.value) + "@" + str(round(token.time, 2))
                if ti < len(self._model_node.marking) - 1:
                    mstr += ", "
                ti += 1
            mstr += "]"
            label = bold_font.render(mstr, True, vis.TUE_RED)
            text_x_pos = self._pos[0] - int(label.get_width()/2)
            text_y_pos = self._pos[1] + self._half_height + vis.LINE_WIDTH + int(label.get_height())
            screen.blit(label, (text_x_pos, text_y_pos))        

    def get_visualisation(self):
        return self.BPMNFlowViz(self)


class BPMNExclusiveSplitGateway(Prototype):
    """
    Generates a composition of SimVar and SimEvent that represents a BPMN exclusive split gateway.
    A choice has a single incoming flow (SimVar) and multiple outgoing flows (SimVar) to choose from.
    A behavior function must be specified that determines which outgoing flow is chosen.
    The behavior function must take a single input parameter, which is the incoming flow.
    The behavior function must return a list of SimToken, which are put on the outgoing flows. For outgoing flows that are not chosen, the behavior function must None. Exactly one SimToken must be returned for each outgoing flow.
    
    
    :param model: the SimProblem to which the event composition must be added.
    :param incoming: a list with exactly one SimVar: a case SimVar.
    :param outgoing: a list with at least two SimVar: a case SimVar.
    :param label: the label on the gateway.
    :param behavior: specifies the choice behavior of the gateway.
    """

    def __init__(self, model, incoming, outgoing, label, behavior):
        super().__init__(model, incoming, outgoing, label)

        if len(incoming) != 1:
            raise TypeError("Gateway " + label + ": must have at exactly one input parameter for cases.")
        if len(outgoing) < 2:
            raise TypeError("Gateway " + label + ": must have at least two output parameter for cases.")
        if not callable(behavior):
            raise TypeError("Gateway " + label + ": the behavior must be a function. (Maybe you made it a function call, exclude the brackets.)")
        if len(inspect.signature(behavior).parameters) != 1:
            raise TypeError("Gateway " + label + ": the behavior function must have at exactly one input parameter for cases.")

        def behavior_encapsulation(c):
            # encapsulates the behavior to check if the behavior function returns a list of SimToken with the same length as the number of outgoing flows.
            result = behavior(c)
            if len(result) != len(outgoing):
                raise TypeError("Gateway " + label + ": the behavior function must return a list of SimToken with the same length as the number of outgoing flows.")
            count_tokens = 0
            for i in range(len(result)):
                if result[i] is not None:
                    if not isinstance(result[i], SimToken):
                        raise TypeError("Gateway " + label + ": the behavior function must return a list of SimToken. However, element " + str(i) + " is not a SimToken.")
                    count_tokens += 1
            if count_tokens != 1:
                raise TypeError("Gateway " + label + ": the behavior function must return a list of SimToken with exactly one SimToken. However, " + str(count_tokens) + " SimTokens were returned.")
            return result
        result = model.add_event(incoming, outgoing, behavior_encapsulation, name=label + "<xor_split>")
        self.add_event(result)

        model.add_prototype(self)

    class BPMNExclusiveSplitGatewayViz(vis.Node):
        def __init__(self, model_node):
            super().__init__(model_node)
        
        def draw(self, screen):
            # draw a pygame diamond shape with TUE_BLUE outline and TUE_LIGHTBLUE fill
            x_pos, y_pos = int(self._pos[0] - self._width/2), int(self._pos[1] - self._height/2)
            pygame.draw.polygon(screen, vis.TUE_LIGHTBLUE, [(x_pos, y_pos + self._half_height), (x_pos + self._half_width, y_pos), (x_pos + self._width, y_pos + self._half_height), (x_pos + self._half_width, y_pos + self._height)])
            pygame.draw.polygon(screen, vis.TUE_BLUE, [(x_pos, y_pos + self._half_height), (x_pos + self._half_width, y_pos), (x_pos + self._width, y_pos + self._half_height), (x_pos + self._half_width, y_pos + self._height)], vis.LINE_WIDTH)
            font = pygame.font.SysFont('Calibri', vis.TEXT_SIZE)

            # # draw a big X inside the diamond
            pygame.draw.line(screen, vis.TUE_BLUE, (x_pos + 0.35*self._width, y_pos + 0.3*self._height), (x_pos + 0.65*self._width, y_pos + 0.7*self._height), 4*vis.LINE_WIDTH)
            pygame.draw.line(screen, vis.TUE_BLUE, (x_pos + 0.35*self._width, y_pos + 0.7*self._height), (x_pos + 0.65*self._width, y_pos + 0.3*self._height), 4*vis.LINE_WIDTH)

            # draw label
            label = font.render(self._model_node.get_id(), True, vis.TUE_BLUE)
            text_x_pos = self._pos[0] - int(label.get_width()/2)
            text_y_pos = self._pos[1] + self._half_height + vis.LINE_WIDTH
            screen.blit(label, (text_x_pos, text_y_pos))

    def get_visualisation(self):
        return self.BPMNExclusiveSplitGatewayViz(self)


class BPMNExclusiveJoinGateway(Prototype):
    """
    Generates a composition of SimVar and SimEvent that represents a BPMN exclusive join gateway.
    A join gateway has multiple incoming flows and one outgoing flow.        
    
    :param model: the SimProblem to which the event composition must be added.
    :param incoming: a list with exactly one SimVar: a case SimVar.
    :param outgoing: a list with at least two SimVar: a case SimVar.
    :param label: the label on the gateway.
    """

    def __init__(self, model, incoming, outgoing, label):
        super().__init__(model, incoming, outgoing, label)

        if len(incoming) < 2:
            raise TypeError("Gateway " + label + ": must have multiple input parameter for cases.")
        if len(outgoing) != 1:
            raise TypeError("Gateway " + label + ": must have exactly one output parameter for cases.")

        self._joiner = model.add_var(label + "_joiner")
        self.add_var(self._joiner)

        for i in range(len(incoming)):
            # create an event, connect it to the incoming flow and the joiner, add it to the model and to self
            result = model.add_event([incoming[i]], [self._joiner], lambda c: [SimToken(c)], name=label + "<joiner" + str(i) + ">")
            self.add_event(result)
        # create an event, connect it to the joiner and the outgoing flow, add it to the model and to self
        resutl = model.add_event([self._joiner], outgoing, lambda c: [SimToken(c)], name=label + "<joiner>")
        self.add_event(resutl)

        model.add_prototype(self)

    class BPMNExclusiveJoinGatewayViz(vis.Node):
        def __init__(self, model_node):
            super().__init__(model_node)
        
        def draw(self, screen):
            # draw a pygame diamond shape with TUE_BLUE outline and TUE_LIGHTBLUE fill
            x_pos, y_pos = int(self._pos[0] - self._width/2), int(self._pos[1] - self._height/2)
            pygame.draw.polygon(screen, vis.TUE_LIGHTBLUE, [(x_pos, y_pos + self._half_height), (x_pos + self._half_width, y_pos), (x_pos + self._width, y_pos + self._half_height), (x_pos + self._half_width, y_pos + self._height)])
            pygame.draw.polygon(screen, vis.TUE_BLUE, [(x_pos, y_pos + self._half_height), (x_pos + self._half_width, y_pos), (x_pos + self._width, y_pos + self._half_height), (x_pos + self._half_width, y_pos + self._height)], vis.LINE_WIDTH)
            font = pygame.font.SysFont('Calibri', vis.TEXT_SIZE)
            bold_font = pygame.font.SysFont('Calibri', vis.TEXT_SIZE, bold=True)

            # # draw a big X inside the diamond
            pygame.draw.line(screen, vis.TUE_BLUE, (x_pos + 0.35*self._width, y_pos + 0.3*self._height), (x_pos + 0.65*self._width, y_pos + 0.7*self._height), 4*vis.LINE_WIDTH)
            pygame.draw.line(screen, vis.TUE_BLUE, (x_pos + 0.35*self._width, y_pos + 0.7*self._height), (x_pos + 0.65*self._width, y_pos + 0.3*self._height), 4*vis.LINE_WIDTH)

            # draw label
            label = font.render(self._model_node.get_id(), True, vis.TUE_BLUE)
            text_x_pos = self._pos[0] - int(label.get_width()/2)
            text_y_pos = self._pos[1] + self._half_height + vis.LINE_WIDTH
            screen.blit(label, (text_x_pos, text_y_pos))
            
            # draw marking
            mstr = "["
            ti = 0
            for token in self._model_node._joiner.marking:
                mstr += str(token.value) + "@" + str(round(token.time, 2))
                if ti < len(self._model_node._joiner.marking) - 1:
                    mstr += ", "
                ti += 1
            mstr += "]"
            label = bold_font.render(mstr, True, vis.TUE_RED)
            text_x_pos = self._pos[0] - int(label.get_width()/2)
            text_y_pos = self._pos[1] + self._half_height + vis.LINE_WIDTH + int(label.get_height())
            screen.blit(label, (text_x_pos, text_y_pos))        

    def get_visualisation(self):
        return self.BPMNExclusiveJoinGatewayViz(self)


class BPMNParallelSplitGateway(SimEvent):
    """
    A SimVar that represents a parallel split gateway.
    It is just a SimEvent with a different visualisation.
    It must have a single incoming place and multiple outgoing places.
    Its behavior takes the data from the source place and passes it to each of its outgoing places.
    """

    def __init__(self, model, incoming, outgoing, label):
        super().__init__(label)

        if len(incoming) != 1:
            raise TypeError("Gateway " + label + ": must have at exactly one input parameter for cases.")
        if len(outgoing) < 2:
            raise TypeError("Gateway " + label + ": must have at least two output parameter for cases.")

        self.set_inflow(incoming)
        self.set_outflow(outgoing)

        self.set_behavior(lambda c: [SimToken(c) for _ in range(len(outgoing))])

        model.add_prototype_event(self)

    class BPMNParallelSplitGatewayViz(vis.Node):
        def __init__(self, model_node):
            super().__init__(model_node)
        
        def draw(self, screen):
            # draw a pygame diamond shape with TUE_BLUE outline and TUE_LIGHTBLUE fill
            x_pos, y_pos = int(self._pos[0] - self._width/2), int(self._pos[1] - self._height/2)
            pygame.draw.polygon(screen, vis.TUE_LIGHTBLUE, [(x_pos, y_pos + self._half_height), (x_pos + self._half_width, y_pos), (x_pos + self._width, y_pos + self._half_height), (x_pos + self._half_width, y_pos + self._height)])
            pygame.draw.polygon(screen, vis.TUE_BLUE, [(x_pos, y_pos + self._half_height), (x_pos + self._half_width, y_pos), (x_pos + self._width, y_pos + self._half_height), (x_pos + self._half_width, y_pos + self._height)], vis.LINE_WIDTH)
            font = pygame.font.SysFont('Calibri', vis.TEXT_SIZE)

            # # draw a big + inside the diamond
            pygame.draw.line(screen, vis.TUE_BLUE, (x_pos + self._half_width, y_pos + 0.2*self._height), (x_pos + self._half_width, y_pos + 0.8*self._height), 4*vis.LINE_WIDTH)
            pygame.draw.line(screen, vis.TUE_BLUE, (x_pos + 0.2*self._width, y_pos + self._half_height), (x_pos + 0.8*self._width, y_pos + self._half_height), 4*vis.LINE_WIDTH)

            # draw label
            label = font.render(self._model_node.get_id(), True, vis.TUE_BLUE)
            text_x_pos = self._pos[0] - int(label.get_width()/2)
            text_y_pos = self._pos[1] + self._half_height + vis.LINE_WIDTH
            screen.blit(label, (text_x_pos, text_y_pos))
            
    def get_visualisation(self):
        return self.BPMNParallelSplitGatewayViz(self)
    

class BPMNParallelJoinGateway(SimEvent):
    """
    A SimVar that represents a parallel join gateway.
    It is just a SimEvent with a different visualisation.
    It must have multiple incoming places and a single outgoing place.
    Its behavior takes the data from one of the source places and passes it to the outgoing place.
    """

    def __init__(self, model, incoming, outgoing, label, behavior=None):
        super().__init__(label)

        if len(incoming) < 2:
            raise TypeError("Gateway " + label + ": must have at least two input parameter for cases.")
        if len(outgoing) != 1:
            raise TypeError("Gateway " + label + ": must have at exactly one output parameter for cases.")

        self.set_inflow(incoming)
        self.set_outflow(outgoing)

        if behavior is not None:
            self.set_behavior(behavior)
        else:
            self.set_behavior(lambda *args: [SimToken(args[0])])
        
        # the transition must be fired for cases with the same case identifier
        self.set_guard(lambda *args: all([args[i][0] == args[0][0] for i in range(1,len(args))]))

        model.add_prototype_event(self)

    class BPMNParallelJoinGatewayViz(vis.Node):
        def __init__(self, model_node):
            super().__init__(model_node)
        
        def draw(self, screen):
            # draw a pygame diamond shape with TUE_BLUE outline and TUE_LIGHTBLUE fill
            x_pos, y_pos = int(self._pos[0] - self._width/2), int(self._pos[1] - self._height/2)
            pygame.draw.polygon(screen, vis.TUE_LIGHTBLUE, [(x_pos, y_pos + self._half_height), (x_pos + self._half_width, y_pos), (x_pos + self._width, y_pos + self._half_height), (x_pos + self._half_width, y_pos + self._height)])
            pygame.draw.polygon(screen, vis.TUE_BLUE, [(x_pos, y_pos + self._half_height), (x_pos + self._half_width, y_pos), (x_pos + self._width, y_pos + self._half_height), (x_pos + self._half_width, y_pos + self._height)], vis.LINE_WIDTH)
            font = pygame.font.SysFont('Calibri', vis.TEXT_SIZE)

            # # draw a big + inside the diamond
            pygame.draw.line(screen, vis.TUE_BLUE, (x_pos + self._half_width, y_pos + 0.2*self._height), (x_pos + self._half_width, y_pos + 0.8*self._height), 4*vis.LINE_WIDTH)
            pygame.draw.line(screen, vis.TUE_BLUE, (x_pos + 0.2*self._width, y_pos + self._half_height), (x_pos + 0.8*self._width, y_pos + self._half_height), 4*vis.LINE_WIDTH)

            # draw label
            label = font.render(self._model_node.get_id(), True, vis.TUE_BLUE)
            text_x_pos = self._pos[0] - int(label.get_width()/2)
            text_y_pos = self._pos[1] + self._half_height + vis.LINE_WIDTH
            screen.blit(label, (text_x_pos, text_y_pos))
            
    def get_visualisation(self):
        return self.BPMNParallelJoinGatewayViz(self)


from simpn.reporters import EventLogReporter
from tempfile import NamedTemporaryFile

class Conformance:
    """
    A class that can calculate the conformance of an event log, produced via the EventLogReporter, to a process model, specified as a SimProblem with BPMN prototypes.
    It has methods to calculate the fitness and the precision metrics.
    Fitness and precision are calculated based on the edit distance similarity between traces from the event log and the process model.
    The edit distance similarity between a trace t from the event log and a trace p from the process model is: the lowest number of insertions, deletions, and substitutions needed to transform t into p, divided by max(length of p, length of t).
    The fitness of a trace t from the event log is the highest edit distance similarity between t and any trace p from the process model.
    The fitness of the event log is the average fitness of all traces in the event log.
    The precision is analogously computed as the average fitness of all traces in the process model with respect to the event log.
    Preconditions:
    - we assume that events are in the log in the order of their occurrence in time.
    - the sim_problem must be in its initial state when the Conformance class is instantiated.
    """

    def __init__(self, sim_problem, event_log, separator=",", case_id_column="case_id", task_column="task", sample_duration=1000):
        """
        Initializes the Conformance class with a SimProblem and an event log.

        :param sim_problem: An instance of SimProblem containing the process model.
        :param event_log: The filename of an event log.
        :param separator: The separator used in the event log file (default is comma).
        :param case_id_column: The name of the column in the event log that contains the case IDs (default is "case_id").
        :param task_column: The name of the column in the event log that contains the task labels (default is "task").
        :param sample_duration: The time the simulator will be run to generate traces to calculate the fitness/ precision over (default is 1000).
        """
        self.INITIAL_STATE = "initial state"
        self.sim_problem = sim_problem
        self.sim_problem.store_checkpoint(self.INITIAL_STATE)  # Store the initial state of the process model
        self.event_log = event_log
        self.separator = separator
        self.case_id_column = case_id_column
        self.task_column = task_column
        self.sample_duration = sample_duration
        self.traces_sampled_from_process = self.generate_traces(duration=sample_duration)
        self.sim_problem.restore_checkpoint(self.INITIAL_STATE)  # Restore the initial state after generating traces
        self.traces_extracted_from_log = self.extract_traces(event_log, separator=separator, case_id_column=case_id_column, task_column=task_column)
    
    def extract_traces(self, event_log, separator=",", case_id_column="case_id", task_column="task"):
        """
        Extracts traces from the event log.
        The result is a dictionary of trace -> frequency of occurrence, where each trace is a list of task labels.

        :return: A dictionary of traces with the frequency of their occurrence.
        """
        with open(event_log, 'r') as file:
            header = file.readline().strip().split(separator)
            case_id_index = header.index(case_id_column)
            task_index = header.index(task_column)

            # A dictionary case_id -> list of tasks
            cases = {}

            for line in file:
                if line.strip():
                    parts = line.strip().split(separator)
                    case_id = parts[case_id_index]
                    task = parts[task_index]

                    if case_id not in cases:
                        cases[case_id] = []
                    cases[case_id].append(task)

            # Convert cases to traces
            traces = {}
            for case_id, tasks in cases.items():
                trace = tuple(tasks)
                if trace not in traces:
                    traces[trace] = 0
                traces[trace] += 1

        return traces

    def generate_traces(self, duration=1000):
        """
        Generates traces from the SimProblem's process model.
        The result is a dictionary of trace -> frequency of occurrence, where each trace is a list of task labels.

        :return: A dictionary of traces with the frequency of their occurrence.
        """

        # We can generate traces by simulating the process model, storing the event log in a temporary file, and then extracting the traces from that log.
        with NamedTemporaryFile() as temp_log:            
            reporter = EventLogReporter(temp_log.name, separator=self.separator)
            self.sim_problem.restore_checkpoint(self.INITIAL_STATE)
            self.sim_problem.simulate(duration, reporter)
            temp_log.flush()

            # Now extract traces from the temporary log file
            traces = self.extract_traces(temp_log.name, separator=self.separator)

            # Clean up the temporary file
            temp_log.close()

            return traces

    def edit_distance_similarity(self, trace1, trace2):
        """
        Calculates the edit distance similarity between two traces.
        The edit distance similarity is defined as the lowest number of insertions, deletions, and substitutions needed to transform trace1 into trace2, divided by max(length of trace1, length of trace2).

        :param trace1: A list representing the first trace.
        :param trace2: A list representing the second trace.
        :return: A float representing the edit distance similarity.
        """
        len1 = len(trace1)
        len2 = len(trace2)

        # Create a distance matrix
        dp = [[0] * (len2 + 1) for _ in range(len1 + 1)]

        for i in range(len1 + 1):
            for j in range(len2 + 1):
                if i == 0:
                    dp[i][j] = j
                elif j == 0:
                    dp[i][j] = i
                elif trace1[i - 1] == trace2[j - 1]:
                    dp[i][j] = dp[i - 1][j - 1]
                else:
                    dp[i][j] = min(dp[i - 1][j] + 1,      # Deletion
                                   dp[i][j - 1] + 1,      # Insertion
                                   dp[i - 1][j - 1] + 1)
        edit_distance = dp[len1][len2]
        max_length = max(len1, len2)
        if max_length == 0:
            return 1.0
        return 1 - (edit_distance / max_length)  # Return similarity as a value

    def calculate_fit(self, from_traces, to_traces):
        """
        Calculates the fitness of a set of traces with respect to another set of traces.
        The fitness is defined as the average similarity of each trace in from_traces to the best matching trace in to_traces.
        :param from_traces: A dictionary of traces with their frequencies.
        :param to_traces: A dictionary of traces with their frequencies.
        :return: A float representing the fitness value.
        """
        fitness_values = []

        for from_trace, from_frequency in from_traces.items():
            max_similarity = 0.0
            for to_trace in to_traces.keys():
                similarity = self.edit_distance_similarity(from_trace, to_trace)
                if similarity > max_similarity:
                    max_similarity = similarity
            fitness_values.append(max_similarity * from_frequency)

        if not fitness_values:
            return 0.0
        return sum(fitness_values) / sum(from_traces.values())

    def calculate_fitness(self):
        """
        Calculates the fitness of the event log with respect to the process model.

        :return: A float representing the fitness value.
        """
        return self.calculate_fit(self.traces_extracted_from_log, self.traces_sampled_from_process)        
    
    def calculate_precision(self):
        """
        Calculates the precision of the event log with respect to the process model.
        :return: A float representing the precision value.
        """
        return self.calculate_fit(self.traces_sampled_from_process, self.traces_extracted_from_log)
